# -*- coding: utf-8 -*-
from ..dataset.serializable import serialize
from ..dataset.deserialize import deserialize
from ..utils.common import lls

import logging
import copy
import base64
import json

# HTTPConnection.debuglevel = 1


import sys
if sys.version_info[0] > 2:
    #from urllib.parse import urlencode
    from urllib.request import urlopen
    from urllib.parse import urlsplit
    from urllib.error import HTTPError
    import requests
    from http.client import HTTPConnection
else:
    #from urllib import urlencode
    from urllib2 import urlopen
    from urlparse import urlsplit
    from urllib2 import HTTPError
    from httplib import HTTPConnection

logger = logging.getLogger(__name__)
#logger.debug('level %d' % (logger.getEffectiveLevel()))
#print('level %d' % (logger.getEffectiveLevel()))

if 0:
    print(logger.propagate)
    print(logger.disabled)
    print(logger.filters)
    print(logger.hasHandlers())
    print(logger.handlers)
    print(logger.level)


commonheaders = {
    'Accept': 'application/json',
    'Accept-Charset': 'utf-8',
    'Accept-Encoding': 'identity',
    'SameSite': 'Strict',
    'Cache-Control': 'no-cache',
    'Connection': 'keep-alive',
    "Content-type": 'application/json'
}


def auth_headers(username=None, password=None, headers=None):
    """Generate HTTP header with auth.

    Parameters
    ----------
    username : str
        user name. Default is `None`.
    password : str
        Password. Set both `username` and `password` to `None` to use
        the values in `pnconfig["node"]`.
    headers : dict
        key-values of HTTP headers, Default is `None` for using
        `jsonio.commonheaders`.

    Examples
    --------
    FIXME: Add docs.

    """
    if headers is None:
        headers = copy.copy(commonheaders)
    if username is None and password is None:
        return headers
    up = bytes((username + ':' + password).encode('ascii'))
    code = base64.b64encode(up).decode("ascii")
    headers.update({'Authorization': 'Basic %s' % (code)})
    return headers


def getJsonObj(url, headers=None, usedict=True, int_key=False, **kwds):
    """ return object from url. url can be http or file.

    :int_key: translate keys from string to int if `True` Raise exception if fails.
    Not using requests.get() as it cannot open file:/// w/o installing
    https://pypi.python.org/pypi/requests-file
    """
    logger.debug('url: %s' % (url))
    stri = urlopen(
        url, timeout=15).read().decode('utf-8')
    #logger.debug('stri ' + stri)
    # print(url,stri)
    # ret = json.loads(stri, parse_float=Decimal)
    # ret = json.loads(stri, cls=Decoder,
    #               object_pairs_hook=collections.OrderedDict)
    ret = deserialize(stri, usedict=usedict, int_key=int_key, **kwds)
    #logger.debug(pformat(ret, depth=6)[:] + '...')
    logger.debug(lls(str(ret), 160))
    return ret


# def jsonREST(url, obj, headers, cmd):
#     """ generic RESTful command handler for POST, PUT, and DELETE.
#     """
#     js = serialize(obj)
#     # %s obj %s headers %s' % (url, obj, headers))
#     logger.debug(url + lls(js, 160))

#     i = 1
#     while True:
#         try:
#             if 0 and sys.version_info[0] > 2:
#                 if cmd == 'POST':
#                     r = requests.post(
#                         url, data=js, headers=headers, timeout=15)
#                 elif cmd == 'PUT':
#                     r = requests.post(
#                         url, data=js, headers=headers, timeout=15)
#                 elif cmd == 'DELETE':
#                     r = requests.post(
#                         url, data=js, headers=headers, timeout=15)
#                 else:
#                     raise ValueError('Bad REST command ' + cmd)
#                 stri = r.text
#             else:
#                 o = urlsplit(url)
#                 u = o.netloc
#                 p = o.path + '?' + o.query + '#' + o.fragment
#                 h = HTTPConnection(u, timeout=15)
#                 h.request(cmd, p, js, headers)
#                 r = h.getresponse()
#                 stri = r.read()
#             # print('ps textx %s\nstatus %d\nheader %s' % (stri, r.status_code, r.headers))
#             break
#         except Exception as e:
#             logger.debug(e)
#             if i >= 1:
#                 logger.error("Give up %s %s after %d tries." % (cmd, url, i))
#                 return None
#             else:
#                 i += 1

#     # ret = json.loads(stri, parse_float=Decimal)
#     # ret = json.loads(stri, cls=Decoder)
#     ret = deserialize(stri)
#     logger.debug(str(ret)[:160] + '...')
#     return ret


def postJsonObj(url, obj, headers):
    """ posts object to url. Returns None if fails.
    """

    return jsonREST(url, obj, headers, 'POST')


def putJsonObj(url, obj, headers):
    """ puts object to url. Returns None if fails.
    """
    return jsonREST(url, obj, headers, 'PUT')


def deleteJsonObj(url, obj, headers):
    """ deletes object from url. Returns None if fails.
    """
    return jsonREST(url, obj, headers, 'DELETE')


def writeJsonObj(o, fn):
    """ Write an object to file fn in json safely
    Return True if successful else False
    """
    for i in range(5):
        try:
            f = open(fn, 'w')
        except OSError:
            logger.warn('unable to open %f for writing. %d' % (fn, i))
    if i == 5:
        logger.error('unable to open %f for writing.' % (fn))
        return False
    json.dump(o, f)
    f.close()
    return True



# -*- coding: utf-8 -*-

from fdi.dataset.deserialize import deserialize
from conftest import SHORT

from test_pal import backup_restore
from serv.test_httppool import getPayload, check_response

from fdi.dataset.product import Product
from fdi.dataset.numericparameter import NumericParameter
from fdi.dataset.stringparameter import StringParameter
from fdi.dataset.eq import deepcmp

from fdi.dataset.deserialize import serialize_args, deserialize_args, deserialize
from fdi.dataset.testproducts import get_demo_product, get_related_product
from fdi.pal.productstorage import ProductStorage
from fdi.pal.productref import ProductRef
from fdi.pal.query import MetaQuery
from fdi.pal.poolmanager import PoolManager, DEFAULT_MEM_POOL
from fdi.pal.httpclientpool import HttpClientPool
from fdi.pns.fdi_requests import safe_client, urn2fdiurl, parse_poolurl
from fdi.utils.common import fullname

import networkx as nx

from flask_httpauth import HTTPBasicAuth
# from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError
import pytest
import os
import json
import time
import urllib
import copy
import time


def setuplogging():
    import logging
    import logging.config
    from . import logdict

    # create logger
    logging.config.dictConfig(logdict.logdict)
    logging.getLogger("requests").setLevel(logging.WARN)
    logging.getLogger("urllib3").setLevel(logging.WARN)
    logging.getLogger("filelock").setLevel(logging.WARN)
    return logging


logging = setuplogging()
logger = logging.getLogger()

logger.setLevel(logging.INFO)
logger.debug('logging level %d' % (logger.getEffectiveLevel()))

test_poolid = __name__.replace('.', '_')


@pytest.fixture(scope="module")
def init_test():
    pass


def chksa(a, k):
    p = 0
    for not_quoted in [False, True]:
        s = serialize_args(*a, **k)
        if p:
            print('s= ', s)
        code, a1, k1 = deserialize_args(s, not_quoted=False)
        assert code == 200
        assert a == a1
        assert k == k1
        s = urllib.parse.unquote(s)
        if p:
            print('S= ', s)
        code, a1, k1 = deserialize_args(s, not_quoted=True)
        assert code == 200
        assert a == a1
        assert k == k1


def test_serialize_args():
    a = ['__foo__', 1, 2, -3, 4.0, 'a', 'b c__d', b'\xde\xad',
         True, None, NumericParameter(42)]
    k = {'__f__': '__g__', 'a': 'r', 'f': 0, 'b': True,
         'k': None, 's': StringParameter('4..2')}
    chksa(a, k)
    a = [[1]]
    k = {}
    chksa(a, k)
    a = []
    k = {'s': 2}
    chksa(a, k)
    a, k = ['__foo', {'3': 4}], dict(d=6)
    chksa(a, k)


def test_gen_url(server):
    """ Makesure that request create corrent url
    """

    aburl, headers = server
    samplepoolname = 'sample_' + test_poolid
    samplepoolurl = aburl + '/' + samplepoolname
    sampleurn = 'urn:' + samplepoolname + ':fdi.dataset.product.Product:10'

    logger.info('Test GET HK')
    got_hk_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='housekeeping', method='GET')
    hk_url = aburl + '/' + samplepoolname + '/hk/'
    assert got_hk_url == hk_url, 'Housekeeping url error: ' + got_hk_url + ':' + hk_url

    logger.info('Test GET classes, urns, tags, webapi url')
    got_classes_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='classes', method='GET')
    classes_url = aburl + '/' + samplepoolname + '/hk/classes'
    assert got_classes_url == classes_url, 'Classes url error: ' + got_classes_url

    got_urns_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='urns', method='GET')
    urns_url = aburl + '/' + samplepoolname + '/hk/urns'
    assert got_urns_url == urns_url, 'Urns url error: ' + got_urns_url

    got_tags_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='tags', method='GET')
    tags_url = aburl + '/' + samplepoolname + '/hk/tags'
    assert got_tags_url == tags_url, 'Housekeeping url error: ' + got_tags_url

    logger.info('Get product url')
    got_product_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='product', method='GET')
    product_url = aburl + '/' + samplepoolname + '/fdi.dataset.product.Product/10'
    assert got_product_url == product_url, 'Get product url error: ' + got_product_url

    logger.info('GET WebAPI  url')
    call = 'tagExists__foo'
    got_webapi_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents=call, method='GET')
    webapi_url = aburl + '/' + samplepoolname + '/' + 'api/' + call
    # '/'
    assert got_webapi_url == webapi_url + \
        '/', 'Get WebAPI url error: ' + got_webapi_url

    logger.info('Post WebAPI url')
    call = 'tagExists__foo'
    got_post_api_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents=call, method='POST')
    post_api_url = aburl + '/' + samplepoolname+'/' + 'api/' + 'tagExists/'
    assert got_post_api_url == post_api_url, 'Post WebAPI url error: ' + \
        got_post_api_url

    logger.info('Post product url')
    got_post_product_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='product', method='POST')
    post_product_url = aburl + '/' + samplepoolname + '/'
    assert got_post_product_url == post_product_url, 'Post product url error: ' + \
                                                     got_post_product_url

    logger.info('Delete product url')
    got_del_product_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='product', method='DELETE')
    del_product_url = aburl + '/urn' + sampleurn
    assert got_del_product_url == del_product_url, 'Delete product url error: ' + \
                                                   got_del_product_url

    logger.info('Delete pool url')
    got_del_pool_url = urn2fdiurl(
        urn=sampleurn, poolurl=samplepoolurl, contents='wipe_pool', method='DELETE')
    del_pool_url = aburl + '/' + samplepoolname + '/wipe'
    assert got_del_pool_url == del_pool_url, 'Delete product url error: ' + got_del_pool_url

    logger.info('Test corrupt request url')
    with pytest.raises(ValueError) as exc:
        err_url = urn2fdiurl(
            urn=sampleurn, poolurl=samplepoolurl, contents='pool', method='GET')
        exc_msg = exc.value.args[0]
        assert exc_msg == 'No such method and contents composition: GET/pool'


def test_tmp_remote_storage(tmp_remote_storage, tmp_prods):

    ps = tmp_remote_storage
    # lit of poolnames
    plst = list(ps.getPools())
    assert len(plst) == 1
    assert plst[0] == 'test_remote_pool'
    pool = ps.getPool(plst[0])

    # empty
    assert pool.getCount() == 0
    assert pool.isEmpty()
    prod_lst = tmp_prods
    refs = []
    for i, p in enumerate(prod_lst):
        ts = 'saved at %d' % time.time_ns()
        refs.append(ps.save(p, tag=ts))
    retrd = []
    for r in refs:
        p = r.product
        retrd.append(p)
        assert p.description.startswith('test-product-')
    assert len(prod_lst) == len(retrd) == pool.getCount()


def est_CRUD_product_by_client(server, local_pools_dir, auth):
    """Client http product storage READ, CREATE, DELETE products in remote
    """
    aburl, headers = server
    client = None
    poolid = test_poolid
    poolurl = aburl + '/' + poolid
    pool = HttpClientPool(poolname=poolid, poolurl=poolurl)
    crud_t(poolid, poolurl, pool, auth, client)


@pytest.fixture(scope=SHORT)
def get_PS_for_CRUD(server, tmp_remote_storage, auth, client):

    logger.info('Init a pstore')

    ps_remote = tmp_remote_storage
    aburl, headers = server
    poolid = ps_remote.getPools()[0]
    pool = ps_remote.getPool(poolid)
    poolurl = pool.poolurl
    if 1:
        if PoolManager.isLoaded(DEFAULT_MEM_POOL):
            PoolManager.getPool(DEFAULT_MEM_POOL).removeAll()
    # this will also register the server side
    pstore = ProductStorage(pool=pool, auth=auth, client=client)
    pstore.register(poolid)
    pool.removeAll()

    assert len(pstore.getPools()) == 1, 'product storage size error: ' + \
                                        str(pstore.getPools())
    assert pstore.getPool(poolid) is not None, 'Pool ' + \
                                               poolid + ' is None.'
    return poolid, poolurl, pool, auth, client, pstore


def test_CRUD_product_by_client(get_PS_for_CRUD):
    """Client http product storage READ, CREATE, DELETE products in remote
    """
    logger.info('start CRUD test')

    poolid, poolurl, pool, auth, client, pstore = get_PS_for_CRUD

    cnt = pool.getCount('fdi.dataset.product.Product')
    assert cnt == 0, 'Local metadata file size is 0'
    assert pool.count == 0

    logger.info('Save data by ' + pool.__class__.__name__)
    x = Product(description='desc test')

    urn = pstore.save(x, geturnobjs=True)
    x.creator = 'httpclient'

    urn2 = pstore.save(x, geturnobjs=True)
    typenm = fullname(x)
    expected_urn = 'urn:' + poolid + ':' + fullname(x)
    assert urn.urn.rsplit(':', 1)[0] == expected_urn, \
        'Urn error: ' + expected_urn
    logger.info('Saved two products.')

    poolpath, scheme, place, pn, un, pw = parse_poolurl(
        poolurl, poolhint=poolid)
    # import http
    # http.client.HTTPConnection.debuglevel = 1

    cnt = pool.getCount(typenm)
    assert cnt == pool.getCount()

    assert cnt == 2 == pool.count

    logger.info('Load product from httpclientpool')
    res = pstore.getPool(poolid).loadProduct(urn2.urn)
    assert res.creator == 'httpclient', 'Load product error: ' + str(res)
    diff = deepcmp(x, res)
    assert diff is None, diff

    logger.info('Search metadata')
    q = MetaQuery(Product, 'm["creator"] == "httpclient"')
    res = pstore.select(q)
    assert len(res) == 1, 'Select from metadata error: ' + str(res)

    logger.info('Delete a product from httpclientpool')
    pstore.getPool(poolid).remove(urn.urn)
    lsn = pstore.getPool(poolid).getCount('fdi.dataset.product.Product')
    assert lsn == 1, 'Delete product local error, len sn : ' + lsn
    logger.info('A load exception message is expected')

    with pytest.raises(NameError):
        res = pstore.getPool(poolid).loadProduct(urn.urn)
    with pytest.raises(NameError):
        res = pstore.getPool(poolid).loadProduct(
            urn.urn.replace('Product', 'this_and_the_last_errors.are.expected'))

    logger.info('Wipe a pool')
    pstore.getPool(poolid).removeAll()
    try:
        assert pool.isEmpty()
        assert pool.count == 0
    except ConnectionError:
        logger.info('Unregistered Pool raises connectionError.')
    tag = '==== Demo Product ===='
    logger.info('test sample demo prod with tag: ' + tag)
    sp = get_demo_product()
    sp.refs['a_ref'] = ProductRef(get_related_product())

    urn = pstore.save(sp, tag=tag)
    print('Sample Prod saved with tag "%s" %s to %s' %
          (tag, urn.urn, pool.poolname))

    logger.info('unregister a pool')
    assert len(pstore.getPools()) == 1, 'product storage size error: ' + \
                                        str(pstore.getPools())
    # unregister locally and remotely
    pstore.unregister(poolid)
    assert len(pstore.getPools()) == 0, 'product storage size error: ' + \
                                        str(pstore.getPools())

    logger.info('Access a non-existing pool and trgger an Error.')
    with pytest.raises(NameError):
        pstore.getPool(poolid + 'NON_EXISTS ') is None


def est_webapi_backup_restore(server):
    """
    """
    aburl, headers = server

    logger.info('Create pools on the server.')
    poolid = test_poolid
    poolurl = aburl + '/' + poolid
    pool = HttpClientPool(poolname=poolid, poolurl=poolurl)

    logger.info('Bacckup/restore a pool on the server.')

    if PoolManager.isLoaded(DEFAULT_MEM_POOL):
        PoolManager.getPool(DEFAULT_MEM_POOL).removeAll()
    # this will also register the server side
    pstore = ProductStorage(pool=pool)
    backup_restore(pstore)


def test_webapi_backup_restore(tmp_remote_storage):
    """
    """
    logger.info('Create pools on the server.')
    pstore = tmp_remote_storage
    logger.info('Bacckup/restore a pool on the server.')
    backup_restore(pstore)


def test_flask_fmt(tmp_pools, server, client,  auth):
    pool, prd, ref, tag = tmp_pools[0]
    aburl, header = server
    prd_urn = ref.urn

    #### /{pool}  /{pool}/  GET  ####

    # server url without slash
    aburl_no_slash = aburl.rstrip('/')
    # get poolname
    x = safe_client(client.get, aburl_no_slash, auth=auth)
    o, code = getPayload(x)
    # check to see if the pool url is malformed
    check_response(o, code=code, failed_case=False)
    # should be a list of names
    poolnames = o['result']
    assert isinstance(poolnames, dict)
    assert pool._poolname in poolnames

    # check to see if the pool url is malformed by checking if
    # the pool's url can be used to retrieve the pool
    from fdi.pal.urn import Urn, parseUrn
    p_pname, p_type, p_index = parseUrn(prd_urn)

    received_poolurls_no_slash = poolnames[pool._poolname].rstrip('/')
    # get pool with products and their urls

    x = safe_client(client.get, received_poolurls_no_slash, auth=auth)
    o, code = getPayload(x)
    # check to see if the pool url is malformed
    check_response(o, code=code, failed_case=False)
    # The product under the correct type and sn has the right tag
    # tags are a string to make btter display.

    # the url is right
    assert o['result']['DataTypes'][p_type][p_index]['url'] == '/'.join(
        (received_poolurls_no_slash, p_type, str(p_index)))
    assert o['result']['DataTypes'][p_type][p_index]['tags'] == str([tag])
    # pool aburl with slash
    # get the pool using url with slash
    received_poolurls_slash = received_poolurls_no_slash+'/'
    x1 = safe_client(client.get, received_poolurls_slash, auth=auth)
    o1, code = getPayload(x1)
    assert o['result'] == o1['result']

    aburl_slash = aburl_no_slash+'/'
    x = safe_client(client.get, aburl_slash, auth=auth)
    o, code = getPayload(x)
    check_response(o, code=code, failed_case=False)
    # should be a list of names
    poolnames = o['result']
    assert isinstance(poolnames, list)
    assert pool._poolname in poolnames


def test_slash(tmp_remote_storage, tmp_prods):
    # add urn
    ps = tmp_remote_storage
    p0_, p11_, p12_, p121, p122, p1221, p12211 = tmp_prods
    urn12211 = ps.save(p12211).urn


def test_hist(tmp_remote_storage, tmp_prods):
    # add urn
    ps = tmp_remote_storage
    p0_, p11_, p12_, p121, p122, p1221, p12211 = tmp_prods
    urn12211 = ps.save(p12211).urn
    p1221.history.add_input(refs={'p1-2-2-1-1': urn12211})
    assert p1221.history.rowCount == 1
    assert p12211.history.rowCount == 0

    urn1221 = ps.save(p1221).urn
    p122.history.add_input(refs={'p1-2-2-1': urn1221})
    assert p121.history.rowCount == 0
    assert p122.history.rowCount == 1

    # get fresh p0, p11, p12 in case they used for something else
    p11 = copy.deepcopy(p11_)
    p12 = copy.deepcopy(p12_)

    urn121 = ps.save(p121).urn
    urn122 = ps.save(p122).urn
    p12.history.add_input(refs={'p1-2-1': urn121, 'p1-2-2': urn122})
    assert p11.history.rowCount == 0
    assert p12.history.rowCount == 2

    urn11 = ps.save(p11).urn
    urn12 = ps.save(p12).urn
    p0 = copy.deepcopy(p0_)
    v = p0.history
    v.add_input(refs={'p1-1': urn11, 'p1-2': urn12})
    assert v.rowCount == 2

    # use ref as labels
    th = v.getTaskHistory(use_name=False, verbose=0)
    pdot = nx.drawing.nx_pydot.to_pydot(th)
    print(pdot.to_string())
    assert len(th.nodes) == 7
    assert len(th.adj) == 7
    assert len(list(th.pred['root'])) == 2

    urn0 = ps.save(p0).urn
    h = p0.history.getTaskHistory(verbose=False)
    h.nodes['root']['product_ref'] = f'"{urn0}"'
    if 1:
        assert len(p11.history['name']) == 0
        assert p12.history['name'][0] == 'p1-2-1'
        assert p12.history['name'][1] == 'p1-2-2'
        assert p122.history['name'][0] == 'p1-2-2-1'
        assert p1221.history['name'][0] == 'p1-2-2-1-1'
        assert nx.is_directed_acyclic_graph(h)

    assert len(h.adj) == 7
    pdot = nx.drawing.nx_pydot.to_pydot(h)
    print(pdot.to_string())


def test_no_auth(tmp_pools, server, client):

    pool, prd, ref, tag = tmp_pools[0]
    aburl, header = server
    prd_urn = ref.urn

    # get pool without auth
    x = safe_client(client.get, aburl, auth=None)
    o, code = getPayload(x)
    # check to see if the pool url is malformed
    check_response(o, code=code, failed_case=False)
    # pool name is found
    assert pool.poolname in o['result']


def test_need_auth(existing_pools, server, client, auth):
    pool = existing_pools[0]
    aburl, header = server

    url = '/'.join((aburl, pool.poolname, 'hk/'))
    # get pool with auth
    x = safe_client(client.get, url, auth=auth)
    cookie = x.cookies
    o, code = getPayload(x)

    # check to see if the pool url is malformed
    check_response(o, code=code, failed_case=False)
    assert o['msg'] == 'OK.'

    # in session
    url = '/'.join((aburl, pool.poolname, 'hk/dTags'))

    x = safe_client(client.get, url, auth=auth)

    o, code = getPayload(x)
    check_response(o, code=code, failed_case=False)
    assert o['msg'] == 'dTags HK data returned OK'



# -*- coding: utf-8 -*-

from .productpool import ManagedPool, PoolNotFoundError, MetaData_Json_Start, MetaData_Json_End
from ..utils.common import pathjoin, trbk, find_all_files
from .dicthk import HKDBS
from .urn import makeUrn, Urn, parseUrn
from ..dataset.deserialize import deserialize


import tarfile
import filelock
from functools import lru_cache
import sys
import shutil
import mmap
import io
import os
from os import path as op
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = str
    from urllib.parse import urlparse, quote, unquote
else:
    PY3 = False
    strset = (str, unicode)
    from urlparse import urlparse, quote, unquote


def wipeLocal(path):
    """
    does the scheme-specific remove-all.

    A new directory at `path` will be created.
    """
    # logger.debug()

    if path == '/':
        raise (ValueError('Do not remove root directory.'))

    try:
        if op.exists(path):
            shutil.rmtree(path)
        os.makedirs(path)
    except Exception as e:
        msg = 'remove-mkdir failed. exc: %s trbk: %s.' % (str(e), trbk(e))
        logger.error(msg)
        raise e


class LocalPool(ManagedPool):
    """ the pool will save all products in local computer.

        :makenew: when the pool does not exist, make a new one (````True```; default) or throws `PoolNotFoundError` (```False```).

    """

    def __init__(self, makenew=True, **kwds):
        """ creates file structure if there isn't one. if there is, read and populate house-keeping records. create persistent files if not exist.
        """
        # print(__name__ + str(kwds))
        self._makenew = makenew  # must preceed setup() in super
        super().__init__(**kwds)

    def setup(self):
        """ Sets up LocalPool interals.

        Make sure that self._poolname and self._poolurl are present.
        """

        if super().setup():
            return True

        real_poolpath = self.transformpath(self._poolname)

        if not op.exists(real_poolpath):
            if self._makenew:
                os.makedirs(real_poolpath)
            else:
                raise PoolNotFoundError('poolname: %r poolurl: %r real_poolpath: %r' % (
                    self._poolname, self._poolurl, real_poolpath))
        self._files = {}
        self._atimes = {}
        self._cached_files = {}

        c, t, u, dTypes, dTags = tuple(self.readHK().values())

        logger.debug('created ' + self.__class__.__name__ + ' ' + self._poolname +
                     ' at ' + real_poolpath + ' HK read.')

        self._classes.update(c)
        self._tags.update(t)
        self._urns.update(u)
        # new ####
        if len(self._dTypes) or len(self._dTags):
            raise ValueError('self._dTypes or self._dTags not empty %s %s' % (
                str(self._dTypes), str(self._dTags)))

        self._dTypes.update(dTypes)
        self._dTags.update(dTags)
        # /new ###

        if any(not op.exists(op.join(real_poolpath, hk+'.jsn'))
               for hk in HKDBS):
            self.writeHK(real_poolpath)
        return False

    def readmmap(self, filename, start=None, end=None, close=False, check_time=False):
        fp = op.abspath(filename)
        if check_time:
            sr = os.stat(fp)
            if fp in self._atimes and (sr.st_mtime_ns <= self._atimes[fp]):
                # file hasnot changed since last time we read/wrote it.
                return None
        try:
            if 1:  # if fp not in self._files or self._files[fp] is None:
                file_obj = open(fp, mode="r+", encoding="utf-8")
                mmap_obj = mmap.mmap(
                    file_obj.fileno(), length=0, access=mmap.ACCESS_READ)
                fo = mmap_obj
            else:
                fo = self._files[fp]
            if start is None:
                js = fo.read()
            else:
                fo.seek(start)
                js = fo.read(end - start)
        except Exception as e:
            msg = 'Error in HK reading. file: %s. exc: %s trbk: %s.' % (
                fp, str(e), trbk(e))
            logging.error(msg)
            raise NameError(msg)
        if 1:  # close:
            fo.close()
            if fp in self._files:
                del self._files[fp]
        else:
            self._files[fp] = fo
        if check_time:
            # save the mtime as the self atime
            self._atimes[fp] = sr.st_mtime_ns
        return js.decode('ascii')

    def readHK(self, hktype=None, serialize_out=False, all_versions=True):
        """
        loads and returns the housekeeping data, or empty `dict` if not found.

        hktype: one of the mappings listed in `dicthk.HKDBS`.
        serialize_out: if True return serialized form. Default is false.
        """
        if hktype is None:
            # some new ####
            hks = HKDBS
        else:
            hks = [hktype]
        fp0 = self.transformpath(self._poolname)

        hk = {}
        for hkdata in hks:
            fp = op.abspath(pathjoin(fp0, hkdata + '.jsn'))
            if op.exists(fp) and (all_versions or ('dT' in fp)):
                js = self.readmmap(fp, check_time=True)
                if js:
                    if serialize_out:
                        r = js
                    else:
                        from ..dataset.deserialize import deserialize
                        r = deserialize(js, int_key=True)
                    self._cached_files[fp] = js
                else:
                    # the file hasnot changed since last time we r/w it.
                    r = self._cached_files[fp] if serialize_out else \
                        self.__getattribute__('_' + hkdata)
            else:
                if serialize_out:
                    r = '{}'  # '{"_STID":"ODict"}'
                else:
                    from ..dataset.odict import ODict
                    r = {}  # ODict()
            hk[hkdata] = r
        logger.debug('HK read from ' + fp0)
        if serialize_out:
            return '{%s}' % ', '.join(('"%s": %s' % (k, v) for k, v in hk.items())) if hktype is None else hk[hktype]
        else:
            return hk if hktype is None else hk[hktype]

    def writeJsonmmap(self, fp, data, serialize_in=True, serialize_out=False, close=False, check_time=False, meta_location=False, **kwds):
        """ write data in JSON from mmap file at fp.

        register the file. Leave file open by default `close`.
        data: to be serialized and saved.
        serialize_out: if True returns contents in serialized form.
        :check_time: to check if file has not been written since we did last time. Default `False`.
        :meta_location: return the start and end offsets of metadata in data JSON.  Default `False`.
        :return:
        int bytes written. If `meta_location` is ```True```, adding int int start and end point offsets of metadata in seriaized data.
        """
        from ..dataset.serializable import serialize

        js = serialize(data, **kwds) if serialize_in else data
        # start = end = None
        if meta_location:
            # locate metadata
            start = js.find(MetaData_Json_Start, 0)
            # make end relative to file start
            end = js.find(MetaData_Json_End, start) + start
            start += len(MetaData_Json_Start)
            end += len(MetaData_Json_End)

        fp = op.abspath(fp)
        if 1:  # fp not in self._files or self._files[fp] is None:
            file_obj = open(fp, mode="w+", encoding="utf-8")
            # with mmap.mmap(file_obj.fileno(), length=0, access=mmap.ACCESS_WRITE) as mmap_obj:
        else:
            file_obj = self._files[fp]
        file_obj.seek(0)

        # file_obj.resize(len(js))
        file_obj.truncate(0)
        file_obj.write(js)
        # file_obj.flush()
        close = 1
        if close:
            file_obj.close()
            if fp in self._files:
                del self._files[fp]
        else:
            self._files[fp] = file_obj
        if check_time:
            # save the mtime as the self atime
            sr = os.stat(fp)
            os.utime(fp, ns=(sr.st_atime_ns, sr.st_mtime_ns))
            self._atimes[fp] = sr.st_mtime_ns
            self._cached_files[fp] = js
        l = len(js)
        logger.debug('JSON saved to: %s %d bytes' % (fp, l))
        if meta_location:
            return l, start, end
        else:
            return l

    def writeHK(self, fp0=None, all_versions=True):
        """ save the housekeeping data to disk
        """

        if fp0 is None:
            fp0 = self._poolpath + '/' + self._poolname
        l = 0
        for hkdata in HKDBS:
            if not all_versions and not 'dT' in hkdata:
                continue
            fp = pathjoin(fp0, hkdata + '.jsn')
            l += self.writeJsonmmap(fp, self.__getattribute__('_' + hkdata),
                                    check_time=True)
        logger.debug('=== '+str(self._dTypes))
        return l

    def setMetaByUrn(self, start, end, urn=None, datatype=None, sn=None):
        """
        Sets the location of the meta data of the specified data to the given URN or a pair of data type and serial number.

        :data: usually serialized Product.
        """
        u, datatype, sn = self.get_missing(
            urn=urn, datatype=datatype, sn=sn, no_check=True)
        # char offset of the start and end points of metadata
        if start >= 0 and end > 0:
            mt = [start, end]
        else:
            mt = [None, None]
        # new ##
        self._dTypes[datatype]['sn'][sn]['meta'] = mt
        # self._urns[u]['meta'] = mt

    @ lru_cache(maxsize=1024)
    def getMetaByUrnJson(self, urn, resourcetype, index):
        """ like `getMetaByUrn` but returns un-deserialized form."""

        urn, datatype, sn = self.get_missing(
            urn=urn, datatype=resourcetype, sn=index)
        # deserialize(prd[start+len(MetaData_Json_Start):end+len(MetaData_Json_End)])
        try:
            # new ##
            # assert tuple(self._urns[urn]['meta']) == \
            # tuple(self._dTypes[datatype]['sn'][sn]['meta'])
            start, end = tuple(self._dTypes[datatype]['sn'][sn]['meta'])
            # start, end = tuple(self._urns[urn]['meta'])
        except KeyError as e:
            msg = f"Trouble with {self._poolname}...['meta']"
            logger.debug(msg)
            raise
        js = self.schematicLoad(resourcetype=datatype,
                                index=sn, start=start, end=end,
                                serialize_out=True)

        return js

    def getCacheInfo(self):
        info = super().getCacheInfo()
        for i in ['getMetaByUrnJson', 'transformpath']:
            info[i] = getattr(self, i).cache_info()

        return info

    def getMetaByUrn(self, urn, resourcetype=None, index=None):
        """ 
        Get all of the meta data belonging to a product of a given URN or a pair of data type and serial number.

        mh: returns an iterator.
        """
        m = self.getMetaByUrnJson(urn, resourcetype=resourcetype,
                                  index=index)
        return deserialize(m)  # self._urns[urn]['meta']

    def doSave(self, resourcetype, index, data, tags=None, serialize_in=True, **kwds):
        """
        does the media-specific saving.

        index: int
        """

        fp0 = self.transformpath(self._poolname)
        fp = pathjoin(fp0, quote(resourcetype) + '_' + str(index))
        try:
            # t0 = time.time()
            l, start, end = self.writeJsonmmap(
                fp, data, serialize_in=serialize_in, close=True,
                meta_location=True, **kwds)

            urn = makeUrn(self._poolname, resourcetype, index)
            self.setMetaByUrn(start, end, urn)

            l += self.writeHK(fp0)
            # print('tl %.8f %9d' % (time.time()-t0, l))
            logger.debug('HK written')
        except IOError as e:
            logger.error('Save failed. exc: %s trbk: %s.' % (str(e), trbk(e)))
            raise e  # needed for undoing HK changes
        return l, start, end

    def doLoad(self, resourcetype, index, start=None, end=None, serialize_out=False):
        """
        does the action of loading.

        serialize_out: if True returns contents in serialized form.
        """

        indexstr = str(index)
        pp = self.transformpath(self._poolname) + '/' + \
            resourcetype + '_' + indexstr
        js = self.readmmap(pp, start=start, end=end, close=True)
        if serialize_out:
            r = js
        else:
            from ..dataset.deserialize import deserialize
            r = deserialize(js)

        return r

    def doRemove(self, resourcetype, index, **kwds):
        """
        does the action of removal of product from pool.
        """

        fp0 = self.transformpath(self._poolname)
        fp1 = [op.abspath(pathjoin(fp0,  quote(r) + f'_{i}'))
               for r, i in zip(resourcetype, index)]
        res = []
        for fp in fp1:
            try:
                if fp in self._files:
                    if self._files[fp]:
                        self._files[fp].flush()
                        self._files[fp].close()
                    del self._files[fp]
                os.unlink(fp)
                self.writeHK(fp0)
                res.append(0)
            except RuntimeError as e:
                msg = f'Remove failed. exc: {e} trbk: {trbk(e)}'
                logger.debug(msg)
                if self.ignore_error_when_delete:
                    res.append(None)
                    continue
                else:
                    raise
        return res

    def doWipe(self):
        """
        does the action of remove-all
        """
        for n, f in self._files.items():
            if f:
                f.flush()
                f.close()
        self._files.clear()
        self._atimes.clear()
        self._cached_files.clear()

        # will leave a newly made pool dir
        wipeLocal(self.transformpath(self._poolname))
        return 0

    def getHead(self, ref):
        """ Returns the latest version of a given product, belonging
        to the first pool where the same track id is found.
        """

        raise (NotImplementedError())

    def backup(self):
        """ make a tarfile string into a string """

        fp0 = self.transformpath(self._poolname)
        logger.info('Making a gz tar file of %s for pool %s.' %
                    (fp0, self._poolname))
        with filelock.FileLock(self.lockpath('r')):
            # Save unsaved changes
            self.writeHK(fp0)
            with io.BytesIO() as iob:
                with tarfile.open(None, 'w|gz', iob) as tf:
                    tar = tf.add(fp0, arcname='.')
                file_image = iob.getvalue()
        return file_image

    def restore(self, tar):
        """untar the input file to this pool and return the file list."""

        with filelock.FileLock(self.lockpath('w')):
            fp0 = self.transformpath(self._poolname)
            self.doWipe()
            with io.BytesIO(tar) as iob:
                with tarfile.open(None, 'r|gz', iob) as tf:
                    tf.extractall(fp0)
            allf = find_all_files(fp0)
            # read into memory
            self._classes, self._tags, self._urns, self._dTypes, self._dTags = tuple(
                self.readHK().values())
        logger.info('Restored from a gz tar file to %s for pool %s. %d files.' %
                    (fp0, self._poolname, len(allf)))

        return allf



# -*- coding: utf-8 -*-

import logging
import copy
from collections import OrderedDict
from fdi.dataset.history import History
from fdi.dataset.copyable import Copyable
from fdi.dataset.eq import deepcmp
from fdi.dataset.listener import EventSender, EventType
from fdi.dataset.abstractcomposite import AbstractComposite

# Automatically generated from fdi/dataset/resources/BaseProduct.yml. Do not edit.

from collections import OrderedDict
from builtins import str
from fdi.dataset.finetime import FineTime


# create logger
logger = logging.getLogger(__name__)


class BaseProduct( AbstractComposite, Copyable, EventSender):
    """ A BaseProduct is the starting point of te whole  product tree, and a generic result that can be passed on between processes.

    In general a Product contains zero or more datasets, history,
    optional reference pointers and metadata some required metadata fields.
    A Product is expected to fully describe itself; this includes
    the way how this product was achieved (its history). As it is
    the result of a process, it should be able to save to and restore
    from an Archive device.

    Many times a Product may contain a single dataset and for this
    purpose the first dataset entry can be accessed by the getDefault()
    method. Note that the datasets may be a composite of datasets
    by themselves.

    A built-in attributes in `Model['metadata']` ("MetaData Parameter" or `MDP`) can be accessed with e.g. ``p.creator``, or p.meta['creator'].value::

        p.creator='foo'
        assert p.creatur=='foo'
        assert p.meta['creator']=='foo'
        p.meta['creator']=Parameter('bar')
        assert p.meta['creator']==Parameter('bar')

    =====
    BaseProduct class schema 1.6 inheriting [None].

Automatically generated from fdi/dataset/resources/BaseProduct.yml on 2022-10-14 23:26:41.001193.

Description:
FDI base class data model

    """

    def __init__(self,
                                  description = 'UNKNOWN',
                 typ_ = 'BaseProduct',
                 level = 'ALL',
                 creator = 'UNKNOWN',
                 creationDate = FineTime(0),
                 rootCause = 'UNKNOWN',
                 version = '0.8',
                 FORMATV = '1.6.0.11',
                 zInfo=None,
                 **kwds):

        # collect MDPs from args-turned-local-variables.
        metasToBeInstalled = copy.copy(locals())
        metasToBeInstalled.pop('__class__', None)
        metasToBeInstalled.pop('kwds', None)
        metasToBeInstalled.pop('self', None)
        metasToBeInstalled.pop('zInfo', None)

        global Model
        # instance variable for Model to be passed down inhritance chains.
        if zInfo is None:
            zInfo = Model

        # must be the first line to initiate meta and zInfo
        # :class: `Attributable` will process MDPs
        super().__init__(zInfo=zInfo, **metasToBeInstalled, **kwds)

        self._history = History()

    @property
    def history(self):
        """ xx must be a property for ``self.xx = yy`` to work in super class after xx is set as a property also by a subclass.
        """
        return self._history

    @history.setter
    def history(self, history):
        self._history = history

    def accept(self, visitor):
        """ Hook for adding functionality to meta data object
        through visitor pattern."""
        visitor.visit(self)

    def targetChanged(self, event):
        pass
        if event.source == self.meta:
            if event.type_ == EventType.PARAMETER_ADDED or \
               event.type_ == EventType.PARAMETER_CHANGED:
                # logger.debug(event.source.__class__.__name__ +   ' ' + str(event.change))
                pass

    def toString(self, level=0,
                 tablefmt='grid', tablefmt1='rst', tablefmt2='psql',
                 extra=False, param_widths=None,
                 matprint=None, trans=True, beforedata='', **kwds):
        """ like AbstractComposite but with history.
        """
        h = self.history.toString(
            level=level,
            tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
            matprint=matprint, trans=trans, **kwds)
        s = super(BaseProduct, self).toString(
            level=level,
            tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
            extra=extra,
            matprint=matprint, trans=trans, beforedata=h, **kwds)
        return s

    string = toString
    txt = toString

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        s = OrderedDict(
            _ATTR_meta=getattr(self, '_meta', None),
            **self.data,
            _ATTR_history=getattr(self, '_history', None),
            _ATTR_listeners=getattr(self, 'listeners', None))
        return s


    @property
    def description(self): pass
    @property
    def type(self): pass
    @property
    def level(self): pass
    @property
    def creator(self): pass
    @property
    def creationDate(self): pass
    @property
    def rootCause(self): pass
    @property
    def version(self): pass
    @property
    def FORMATV(self): pass
    pass

# Data Model specification for mandatory components
_Model_Spec = {
    'name': 'BaseProduct',
    'description': 'FDI base class data model',
    'parents': [
        None,
        ],
    'schema': '1.6',
    'metadata': {
        'description': {
                'id_zh_cn': '描述',
                'data_type': 'string',
                'description': 'Description of this product',
                'description_zh_cn': '对本产品的描述。',
                'default': 'UNKNOWN',
                'valid': '',
                'typecode': 'B',
                },
        'type': {
                'id_zh_cn': '产品类型',
                'data_type': 'string',
                'description': 'Product Type identification. Name of class or CARD.',
                'description_zh_cn': '产品类型。完整Python类名或卡片名。',
                'default': 'BaseProduct',
                'valid': '',
                'typecode': 'B',
                },
        'level': {
                'id_zh_cn': '产品xx',
                'data_type': 'string',
                'description': 'Product level.',
                'description_zh_cn': '产品xx',
                'default': 'ALL',
                'valid': '',
                'typecode': 'B',
                },
        'creator': {
                'id_zh_cn': '本产品生成者',
                'data_type': 'string',
                'description': 'Generator of this product.',
                'description_zh_cn': '本产品生成方的标识，例如可以是单位、组织、姓名、软件、或特别算法等。',
                'default': 'UNKNOWN',
                'valid': '',
                'typecode': 'B',
                },
        'creationDate': {
                'id_zh_cn': '产品生成时间',
                'fits_keyword': 'DATE',
                'data_type': 'finetime',
                'description': 'Creation date of this product',
                'description_zh_cn': '本产品生成时间',
                'default': 0,
                'valid': '',
                'typecode': None,
                },
        'rootCause': {
                'id_zh_cn': '数据来源',
                'data_type': 'string',
                'description': 'Reason of this run of pipeline.',
                'description_zh_cn': '数据来源（此例来自鉴定件热真空罐）',
                'default': 'UNKNOWN',
                'valid': '',
                'typecode': 'B',
                },
        'version': {
                'id_zh_cn': '版本',
                'data_type': 'string',
                'description': 'Version of product',
                'description_zh_cn': '产品版本',
                'default': '0.8',
                'valid': '',
                'typecode': 'B',
                },
        'FORMATV': {
                'id_zh_cn': '格式版本',
                'data_type': 'string',
                'description': 'Version of product schema and revision',
                'description_zh_cn': '产品格式版本',
                'default': '1.6.0.11',
                'valid': '',
                'typecode': 'B',
                },
        },
    'datasets': {
        },
    }

Model = copy.deepcopy(_Model_Spec)

MdpInfo = Model['metadata']



# -*- coding: utf-8 -*-

from ..utils.common import mstr, bstr
from .listener import MetaDataListener
from .serializable import Serializable
from .typed import Typed
from .odict import ODict
from .attributable import Attributable
from .abstractcomposite import AbstractComposite
from .datawrapper import DataContainer, DataWrapper
from .eq import DeepEqual
from .copyable import Copyable
from .annotatable import Annotatable


from collections import OrderedDict
import logging
import sys

if sys.version_info[0] + 0.1 * sys.version_info[1] >= 3.3:
    PY33 = True
    from collections.abc import Container, Sequence, Mapping
    seqlist = Sequence
    maplist = Mapping
else:
    assert 0, 'python 3'
    PY33 = False
    from .collectionsMockUp import ContainerMockUp as Container
    from .collectionsMockUp import SequenceMockUp as Sequence
    from .collectionsMockUp import MappingMockUp as Mapping
    seqlist = (tuple, list, Sequence, str)
    # ,types.XRangeType, types.BufferType)
    maplist = (dict, Mapping)

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

# from .composite import


class Dataset(Attributable, DataContainer, Serializable, MetaDataListener):
    """ Attributable and annotatable information data container
    that can be be part of a Product.

    developer notes:
    The intent is that developers do not derive from this interface

    directly. Instead, they should inherit from one of the generic
    datasets that this package provides:

    mh: GenericDataset, UnstructuredDataset
    ArrayDataset.
    TableDataset or
    CompositeDataset.
    """

    def __init__(self, **kwds):
        """
        Parameter
        ---------

        Returns
        -------
        """
        super().__init__(**kwds)

    def accept(self, visitor):
        """ Hook for adding functionality to object
        through visitor pattern.
        Parameter
        ---------

        Returns
        -------
        """
        visitor.visit(self)

    def toString(self, level=0,
                 tablefmt='grid', tablefmt1='simple', tablefmt2='psql',
                 param_widths=None, width=0, matprint=None, trans=True,
                 heavy=True, center=-1, **kwds):
        """ matprint: an external matrix print function
        trans: print 2D matrix transposed. default is True.
        Parameter
        ---------

        Returns
        -------

        """
        cn = self.__class__.__name__
        if level > 1:
            return cn + \
                '{ %s, description = "%s", meta = %s }' % \
                (str(self.data), str(self.description), self.meta.toString(
                    tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                    level=level, width=width, param_widths=param_widths,
                    matprint=matprint, trans=trans, heavy=heavy, **kwds))
        html = 'html' in tablefmt.lower() or 'html' in tablefmt2.lower()
        br = '<br>' if html else '\n'

        s, last = make_title_meta_l0(self, level=level, width=width, heavy=heavy,
                                     tablefmt=tablefmt, tablefmt1=tablefmt1,
                                     tablefmt2=tablefmt2, center=center,
                                     param_widths=param_widths,
                                     html=html, excpt=['description'])
        width = len(last) - 1
        if html:
            d = '<center><u>%s</u></center>\n' % 'DATA'
        else:
            d = 'DATA'.center(width) + '\n' + '----'.center(width) + '\n'

        d += bstr(self.data, level=level, heavy=heavy, center=center,
                  tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                  yaml=True, param_widths=param_widths, html=html,
                  **kwds) if matprint is None else \
            matprint(self.data, level=level, trans=False, headers=[],
                     tablefmt2='rst', heavy=heavy,
                     **kwds)
        return f'{s}\n{d}\n{last}\n'

    string = toString
    txt = toString

    def __getstate__(self):
        """ Can be encoded with serializableEncoder.


        Parameter
        ---------

        Returns
        -------
        """

        s = OrderedDict(description=self.description,
                        meta=self._meta,
                        data=self.data)
        return s


def make_title_meta_l0(self, level=0,
                       heavy=True, center=0, html=False, **kwds):
    """ make toString title and metadata.

    :heavy: use bold symbols for separaters.
    :center: 0 for no centering;  -1 for centering with metadata table; other  for ``str.center(<center>``.
    """
    # title
    cn = self.__class__.__name__
    desc = self.meta.get('description', '')
    if desc:
        desc = desc.value
    t = ('*** <b>%s (%s)</b> ***' if html else '*** %s (%s) ***') % (cn, desc)
    tw = len(t)
    # make the table and find out the width first
    table = mstr(self._meta, level=level, html=html, **kwds)
    if center and not html:
        # max separation between consequitive '\n' s
        if center == -1:
            width = max(len(x) for x in table[:600].split('\n'))
        else:
            width = center
    else:
        width = tw

    if heavy:
        # beginning of a stand-alone dataset
        l = ('*' if level else '*') * tw
    else:
        # beginning of a dataset that is a component of a another
        l = '_' * tw

    br = '<br>' if html else '\n'
    if center:
        if html:
            t = '<center>%s</center>' % t
            l = '<center>%s</center>' % l
            m = '<center><u>%s</u></center>' % 'META'
        else:
            t = t.center(width)
            l = l.center(width)
            m = 'META'.center(width)
    else:
        m = 'META%s----%s' % (br, br)
    t += '\n'
    l += '\n'

    if center:
        s = t if level else (l + t + m)
    else:
        s = l + t + ('' if level else l + m)
    s += table
    if html:
        last = '<hr size="3">' if heavy else '<hr>'
    else:
        last = "=" * width if heavy else "~" * width
    last += br

    return s, last


class GenericDataset(Dataset, Typed, DataWrapper):
    """ mh: Contains one typed data item with a unit and a typecode.
    """

    def __init__(self, **kwds):
        """
        """
        super().__init__(**kwds)  # initialize data, meta, unit

    def __iter__(self):
        for x in self.getData():
            yield x


class CompositeDataset(MetaDataListener, AbstractComposite):
    """  An CompositeDataset is a Dataset that contains zero or more
    named Datasets. It allows to build arbitrary complex dataset
    structures.

    It also defines the iteration ordering of its children, which is
    the order in which the children were inserted into this dataset.
    """

    def __init__(self, **kwds):
        """
        Parameter
        ---------

        Returns
        -------

        """
        super(CompositeDataset, self).__init__(
            **kwds)  # initialize _sets, meta, unit

    def __getstate__(self):
        """ Can be encoded with serializableEncoder

        Parameter
        ---------

        Returns
        -------

        """
        return OrderedDict(  # description=self.description,
            _ATTR_meta=self._meta,
            **self._data)



import os
from setuptools import setup, find_packages

# https://pythonhosted.org/an_example_pypi_project/setuptools.html
# https://code.tutsplus.com/tutorials/how-to-write-package-and-distribute-a-library-in-python--cms-28693
#

# Version info -- read without importing
# https://github.com/aio-libs/aiohttp-theme/blob/master/setup.py
_locals = {}
with open('fdi/_version.py') as fp:
    exec(fp.read(), None, _locals)
version = _locals['__version__']

pkgd = os.path.dirname(__file__)


def read(fname):
    return open(os.path.join(pkgd, fname), encoding='utf-8').read()


setup(
    name="fdi",
    version=version,
    author="Maohai Huang",
    author_email="mhuang@earth.bao.ac.cn",
    description=("Flexible Data Integrator"),
    license="LGPL v3",
    keywords="dataset metadata processing product context serialization server URN RESTful API HCSS",
    url="http://mercury.bao.ac.cn:9006/mh/fdi",
    packages=find_packages(exclude=['tests', 'tmp', 'docs']),
    include_package_data=True,
    long_description=read('README.rst'),
    long_description_content_type="text/x-rst",
    python_requires=">=3.6",
    install_requires=[
        'requests>=2.23.0',
        'filelock>=3.0.12',
        'ruamel.yaml<0.17',
        'tabulate-expwidth>=0.9.0.1',
        # for actual use
        #'tabulate @ file://%s/resources/tabulate-0.8.10-py3-none-any.whl' % pkgd,
        # not uploadable for pypi
        #'tabulate @ git+https://github.com/mhuang001/python-tabulate.git@r1',
        #'tabulate @ git+http://mercury.bao.ac.cn:9006/mirrors/tabulate.git@r1',
        #'tabulate @ git+ssh://git@mercury.bao.ac.cn:9005/mirrors/tabulate.git@r1',
        'cwcwidth>=0.1.5',
        'paho-mqtt>=1.6.1',
        'jsonschema>=3.2.0',
        'xmltodict>=0.12.0',
        'jsonpath-ng>=1.5.3',
        'pypng',
        'networkx>=2.8.1',
        'pydot>=1.4.2'
    ],
    entry_points={'console_scripts': [
        'yaml2python=fdi.dataset.yaml2python:main',
        'fdi-getconfig=fdi.utils.getconfig:main',
    ]},
    setup_requires=[],
    tests_require=['pytest', 'pytest-cov', ],
    extras_require={
        'DEV': [
            'setuptools>=43.0.0',
            'wheel>=0.32.1',
            'pytest>=5.4.1',
            'pytest-cov',
            'remote-pdb',
            'asyncio',
            'aiohttp'
        ],
        'SERV': [
            'requests == 2.27.1',
            'urllib3 == 1.26.13',
            'Flask_HTTPAuth >= 4.1.0',
            'Flask>=2.2.2',
            'Werkzeug[watchdog]',
            'uwsgi>=2.0.20',
            'flasgger>=0.9.5',
        ],
        'SCI': [
            'astropy'
        ],
        'PUB': [
            'sphinx>=4.4.0',
            'sphinx_rtd_theme',
            'alabaster>=0.7.12',
            'sphinx-copybutton>=0.3.0',
            'twine>=3.3.0'
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Framework :: Flask",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ],
)



# -*- coding: utf-8 -*-

from .datawrapper import DataContainer
from .odict import ODict
from .serializable import Serializable
from .invalid import INVALID

from collections import UserDict
from collections.abc import MutableMapping
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

# from .listener import DatasetEventSender, DatasetListener
# from .metadata import DataWrapperMapper

# order of Container, Sized, Iterator must be the same as in DataContaier!


# class Composite(DeepEqual, UserDictAdapter, Serializable):  # ):

class Composite(DataContainer, Serializable, MutableMapping):
    """ A container of named Datasets.

    This container can hold zero or more datasets, each of them
    stored against a unique name. The order of adding datasets
    to this Composite is important, that is:
    the keySet() method will return a set of labels of the datasets
    in the sequence as they were added.
    Note that replacing a dataset with the same name,
    will keep the order.

    :class:`DeepEqual` must stay to the left of :class:`MutableMapping` so its `__eq__` will get to run. `Serializable` becomes a parent for having `__setstate__`.

    :data: default `None` will init with a `dict`.
    """

    def __init__(self, data=None, **kwds):
        """

        Parameters
        ----------

        Returns
        -------

        """
        # pass data to parent
        super().__init__({} if data is None else data, **kwds)

    def containsKey(self, name):
        """ Returns true if this map contains a mapping for
        the specified name.

        Parameters
        ----------

        Returns
        -------

        """
        return self.__contains__(name)

    def get(self, name, default=INVALID):
        """ Returns the dataset to which this composite maps the
        specified name.

        If the attrbute does not exist and `default` unspecified, raise a KeyError.

        Parameters
        ----------
        :default: assigne to None or anything for missing `name`.
        Returns
        -------

        """
        if name in self.data:
            return self.data[name]
        elif default is INVALID:
            raise KeyError(name + 'not found')
        else:
            return default

    def set(self, name, dataset):
        """ Associates the specified dataset with the specified key
        in this map(optional operation). If the map previously
        contained a mapping for this key, the old dataset is
        replaced by the specified dataset.
        this composite does not permit null or empty keys.

        Parameters
        ----------

        Returns
        -------

        """
        self._data[name] = dataset

    def __getitem__(self, name):
        """
        Parameters
        ----------

        Returns
        -------

        """

        return self._data[name]

    def __delitem__(self, name):
        """
        Parameters
        ----------

        Returns
        -------

        """
        del self.data[name]

    def __setitem__(self, name, dataset):
        """
        Parameters
        ----------

        Returns
        -------

        """

        if 'refs' in self._data and name != 'refs' and not (name in self._data):
            refs = self._data.pop('refs')
            self._data[name] = dataset
            self._data['refs'] = refs
        else:
            # this must call `set` which can be overridden by subclasses
            self.set(name, dataset)

    def getSets(self):
        """ Provide access to the Map < String, Dataset > .
        mh: api from CompositeDataset

        Parameters
        ----------

        Returns
        -------

        """
        return self.data

    def isEmpty(self):
        """ Returns true if this map contains no key - value mappings.
        Parameters
        ----------

        Returns
        -------

        """
        return len(self.data) == 0

    def keySet(self):
        """ Returns a list view of the keys contained in this composite. 

        Parameters
        ----------

        Returns
        -------

        """
        return list(self.data.keys())

    # convenience method name.
    getDatasetNames = keySet

    def remove(self, name):
        """ Removes the mapping for this name from this composite.
        mh: returns None if name is None or item does not exist.

        Parameters
        ----------

        Returns
        -------

        """
        if name == '' or name is None or name not in self.data:
            logger.debug('Cannot remove non-exist item \'' + name + "'")
            return None
        del self._data[name]

    def size(self):
        """ Returns the number of key - value mappings in this map. 
        Parameters
        ----------

        Returns
        -------

        """
        return len(self.data)

    def getDefault(self):
        """ Convenience method that returns the first dataset \
        belonging to this composite. """

        ls = list(self.values())
        return ls[0] if len(ls) > 0 else None

    def __iter__(self):
        for i in self.data:
            yield i

    def __repr__(self):
        return self.__class__.__name__ + '(' + (self.data.__repr__() if hasattr(self, 'data') else 'None') + ')'

    def toString(self, level=0, matprint=None, trans=True, **kwds):
        """

        Parameters
        ----------

        Returns
        -------

        """
        o = ODict(self.data)
        return self.__class__.__name__ + \
            o.toString(level=level,
                       matprint=matprint, trans=trans, **kwds)

    string = toString
    txt = toString


class UserDictAdapter(UserDict):
    """ Adapter class to make UserDict cooperative to multiple inheritance and take data keyword arg.

    REf. https://rhettinger.wordpress.com/2011/05/26/super-considered-super/
    """

    def __init__(self, data=None, *args, **kwds):
        """

        Parameters
        ----------
        :data: initialize UserDict.

        Returns
        -------

        """
        # pass data to UserDict as the first positional arg
        super().__init__(data, *args, **kwds)



# -*- coding: utf-8 -*-

# https://flask.palletsprojects.com/en/2.2.x/tutorial/views/

import functools

from flask import (
    Blueprint, flash, g, redirect, render_template, request, session, url_for
)
from werkzeug.security import check_password_hash, generate_password_hash

from .db import get_db

user = Blueprint('user', __name__)


@user.before_app_request
def load_logged_in_user():
    user_id = session.get('user_id')

    if user_id is None:
        g.user = None
    else:
        g.user = get_db().execute(
            'SELECT * FROM user WHERE id = ?', (user_id,)
        ).fetchone()


def login_required(view):
    @functools.wraps(view)
    def wrapped_view(**kwargs):
        if g.user is None:
            return redirect(url_for('auth.login'))

        return view(**kwargs)

    return wrapped_view



# -*- coding: utf-8 -*-

from ruamel.yaml import YAML
import os.path as op
import sys

yaml = YAML()

d = op.join(op.abspath(op.dirname(__file__)),
            '../../share/svom/products/vt/resources/VT_frame_common.yml')
with open(d) as f:
    dic = yaml.load(f.read())

for n, att in dic['metadata'].items():
    line = '\t|'.join([n,
                       str(att['default']),
                       att.get('description_zh_cn', ''),
                       att.get('description', ''),
                       ])
    print(line)

sys.exit(0)

if 1:
    import requests
    from fdi.pns.jsonio import commonheaders, auth_headers

auth = auth_headers('foo', 'bar', commonheaders)

url = 'http://127.0.0.1:5000/fdi/v0.10'

x = requests.get(url)

print(x.status_code, x.text)


# url = 'http://127.0.0.1:5000/fdi/v0.10/pools'
# x = requests.get(url, auth=auth_headers)
# print(x.status_code, x.text)



# -*- coding: utf-8 -*-
from .taggable import Taggable
from .urn import Urn, parseUrn, makeUrn
from fdi.dataset.odict import ODict

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

# List of Housekeeping DBs
# some new ####
HKDBS = ['classes', 'tags', 'urns', 'dTypes', 'dTags']


def get_missing(self, urn, datatype, sn, no_check=False):
    """ make URN(s) if datatype and sn(s) are given and vice versa.

    Parameters
    ----------

    no_check: bool
        Do not Check if `datatype` and `sn` are in the pool's HK.
        Default is `False`
    Return
    ------
    tuple
    str, str, int
        Refer tp `parseUrn`.

    Raises
    ------
    ValueError if urn not found or not from this pool.
    KeyError if datatype does not exist.
    IndexError if sn does not exist.
    """
    if urn is None and datatype is None and sn is None:
        return None, None, None

    if datatype is None or sn is None and urn is not None:
        # new ###
        poolname, datatype, sn = parseUrn(urn, int_index=True)
    else:
        # datatype+sn takes priority over urn
        urn = makeUrn(self._poolname, datatype, sn)

    u = urn.urn if issubclass(urn.__class__, Urn) else urn
    # new ###
    if not no_check:
        dat = datatype
        sns = sn
        if not issubclass(sn.__class__, (list, tuple)):
            dat = [datatype]
            sns = [sn]
            for d, s in zip(dat, sns):
                if d not in self._dTypes:
                    raise KeyError(
                        f'{d} not found in pool {self._poolname}')
                if s not in self._dTypes[d]['sn']:
                    raise IndexError('%s:%d not found in pool %s.' %
                                     (d, s, self._poolname))
    # /new ###
    if 0:
        if u not in self._urns:
            raise ValueError(urn + ' not found in pool ' + self._poolname)
    return u, datatype, sn


def add_tag_datatype_sn(tag, datatype, sn, dTypes=None, dTags=None):
    """Static function to  add a tag to datatype-sn to pool fmt 2.0

    Parameters
    ----------
    tag : str
        A tag. Multiple tags have to make multiple calls. `None` and empty tags are ignored.
    datatype : str
        The class name of the data item, new or existing.
    sn : int
        The serial number in integer.
    dTypes : dict
        the first nested mapping of pool fmt 2.
    dTags : dict
        the tag mapping of pool fmt 2.

    """
    if not tag:
        return
    snt = dTypes[datatype]['sn'][sn]['tags']
    if tag not in snt:
        snt.append(tag)
    # dTags saves datatype:sn
    typ = datatype
    if tag not in dTags:
        dTags[tag] = {}
    t = dTags[tag]
    if typ not in t:
        t[typ] = [str(sn)]
    else:
        t[typ].append(str(sn))


class DictHk(Taggable):
    """
    Definition of services provided by a product storage supporting versioning.
    """

    def __init__(self, **kwds):
        super(DictHk, self).__init__(**kwds)
        # {tag->{'urns':[urn]}
        self._tags = dict()
        # {urn->{'tags':[tag], 'meta':meta}}
        self._urns = dict()
        # new ###
        self._dTypes = dict()
        self._dTags = dict()

    def getTags(self, urn=None, datatype=None, sn=None):
        """ 
        Get all of the tags that map to a given URN or a pair of data type and serial number.

        Get all known tags if input arenot specified.
        mh: returns an iterator.

        If datatype and sn are given, use them and ignore urn.
        """

        urn, datatype, sn = self.get_missing(
            urn=urn, datatype=datatype, sn=sn)
        if urn is None:
            return self._dTags.keys()

        # new ###
        if 0:
            assert self._urns[urn]['tags'] == self._dTypes[datatype]['sn'][sn]['tags']
            return self._urns[urn]['tags']

        return self._dTypes[datatype]['sn'][sn]['tags']

    get_missing = get_missing

    def getTagUrnMap(self):
        """
        Get the full tag->urn mappings.

        mh: returns an iterator
        csdb: csdb/v1/storage/tag?tag=tag1,tag2
        """
        # new ###
        return self._dTags

        if 0:
            return zip(self._tags.keys(), map(lambda v: v['urns'], self._value()))

    def getUrn(self, tag):
        """
        Gets the URNs corresponding to the given tag.

        Returns an empty list if `tag` is not `None` and does not exist.
        curl -X GET "http://123.56.102.90:31702/csdb/v1/storage/info?urns=urn%3Apoolbs%3A20211018%3A1" -H "accept: */*"

        """
        if 0:
            if tag not in self._tags:
                return []
        if tag not in self._dTags:
            return []
        # new ###
        if 0:
            assert list(self._tags) == list(self._dTags)
            assert list(self._tags[tag]['urns']) == list(':'.join(
                ['urn', self._poolname, cl, sn]) for cl in self._dTags[tag] for sn in self._dTags[tag][cl])
            return self._tags[tag]['urns']
        # datatype:[sn] -> [urn:poolname:datatype:sn]
        # return ['urn:%s:%s' % (self._poolname, t) for t in self._dTags[tag]]
        t = self._dTags[tag]
        pn = self._poolname
        return list(':'.join(['urn', pn, cl, sn]) for cl in t for sn in t[cl])

    def getUrnObject(self, tag):
        """
        Gets the URNobjects corresponding to the given tag.
        Returns an empty list if `tag` does not exist.
        """

        if 0:
            assert list(self._tags[tag]['urns']) == list(self._dTags[tag])
            return [Urn(x) for x in self._tags[tag]['urns']]
        return [Urn(x) for x in self._dTags[tag]]

    def getAllUrns(self):
        """ Returns a list of all URNs in the pool."""
        res = []
        poolname = self.poolname
        for cls, v in self._dTypes.items():
            res.extend(f'urn:{poolname}:{cls}:{sn}' for sn in v['sn'])
        return res

    def removekey(self, key, thecontainer, thename, cross_ref_map, othername):
        """
        Remove the given key from `the map` and the counterpart key in the correponding `cross_referencing map`.
        """
        vals = thecontainer.pop(key, [])
        # remove all items whose v is key in cross_ref_map
        for val in vals[othername]:
            cross_ref_map[val][thename].remove(key)
            # if we have just removed the last key, remove the empty dict
            if len(cross_ref_map[val][thename]) == 0:
                cross_ref_map[val].pop(thename)
                # if this caused the cross_ref_map[val] to be empty, remove the empty dict
                if len(cross_ref_map[val]) == 0:
                    cross_ref_map.pop(val)

    def removeTag(self, tag):
        """
        Remove the given tag from the tag and urn maps.
        # TODO in CSDB
        """
        # new ##
        clsn_sns = self._dTags.pop(tag)
        for datatype, sns in clsn_sns.items():
            for sn in sns:
                sn = int(sn)
                ts = self._dTypes[datatype]['sn'][sn]['tags']
                if tag in ts:
                    ts.remove(tag)
                    if len(tags) == 0:
                        del ts
                else:
                    logger.warning('tag %s missing from %s:%s:%s.' %
                                   (tag, self._poolname, datatype, sn))
        if 0:
            self.removekey(tag, self._tags, 'tags', self._urns, 'urns')
        # new ##
        # assert list(self._tags) == list(self._dTags)

    def removeUrn(self, urn=None, datatype=None, sn=None):
        """
        Remove the given urn (or a pair of data type and serial number) from the tag and urn maps.

        Only changes maps in memory, not to write on disk here.
        """
        u, datatype, sn = self.get_missing(
            urn=urn, datatype=datatype, sn=sn,
            no_check=False)
        # new ##
        from .productpool import ProductPool
        dats, sns, alist = ProductPool.vectorize(datatype, sn)
        for d, s in zip(dats, sns):
            # if d not in self._dTypes:
            #     if self.ignore_error_when_delete:
            #         return -1
            #     else:
            #         raise ValueError(f'{d} not found in {self.poolname}.')
            _snd = self._dTypes[d]['sn']
            if s not in _snd:
                msg = f'{s} not found in pool {self.getId()}.'
                if not self.ignore_error_when_delete:
                    raise IndexError(msg)
                else:
                    logger.debug(msg)
                    continue
            if 'tags' in _snd[s]:
                for tag in _snd[s]['tags']:
                    if tag in self._dTags:
                        self._dTags[tag][d].remove(str(s))
                        if len(self._dTags[tag][d]) == 0:
                            del self._dTags[tag][d]
                            if len(self._dTags[tag]) == 0:
                                del self._dTags[tag]
                    else:
                        logger.warning('tag %s missing from %s.' %
                                       (tag, self._poolname))
                else:
                    if 0:
                        logger.warning('tag %s missing from %s:%s:%s.' %
                                       (tag, self._poolname, d, s))
            _snd.pop(s)
            if len(_snd) == 0:
                del self._dTypes[d]
            # /new ##

            if 0:
                self.removekey(u, self._urns, 'urns', self._tags, 'tags')
            # new ##
            # assert s not in self._dTypes[d]['sn']

    def setTag(self, tag, urn=None, datatype=None, sn=None):
        """
        Sets the specified tag to the given URN or a pair of data type and serial number.

        # TODO in CSDB
        """
        u, datatype, sn = self.get_missing(
            urn=urn, datatype=datatype, sn=sn, no_check=True)

        if 0:
            self._urns[u]['tags'].append(tag)

            if tag in self._tags:
                self._tags[tag]['urns'].append(u)
            else:
                self._tags[tag] = {'urns': [u]}

        # new ###
        add_tag_datatype_sn(tag, datatype, sn,
                            dTypes=self._dTypes, dTags=self._dTags)
        if 0:
            snt = self._dTypes[datatype]['sn'][sn]['tags']
            if tag not in snt:
                snt.append(tag)
            # dTags saves datatype:sn
            _, typ, sn = tuple(u.rsplit(':', 2))
            if tag not in self._dTags:
                self._dTags[tag] = {}
            t = self._dTags[tag]
            if typ not in t:
                t[typ] = [sn]
            else:
                t[typ].append(sn)

    def tagExists(self, tag):
        """
        Tests if a tag exists.

        """
        # new ##
        if 0:
            assert (tag in self._dTags) == (tag in self._tags)
            return tag in self._tags
        return tag in self._dTags



# -*- coding: utf-8 -*-
from .odict import ODict
from .quantifiable import Quantifiable
from .eq import DeepEqual
from .copyable import Copyable
from .annotatable import Annotatable

from collections.abc import Container, Sized
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class DataContainer(Annotatable, Copyable, DeepEqual, Container):
    """ A DataContainer is a composite of data and description.

    mh: note that There is no metadata.
    Implemented partly from AbstractDataWrapper.
    """

    def __init__(self, data=None, **kwds):
        """
        data: a Container. Default is None.

        Parameters
        ----------

        Returns
        -------

        """
        #print(__name__ + str(kwds))

        if 1:  # data is None:  # or issubclass(data.__class__, Container):
            super().__init__(**kwds)  # DataContainer
            self.setData(data)
        else:
            raise TypeError('DataContainer needs a Container to initialize, not ' +
                            type(data).__name__)

    @property
    def data(self):
        """
        Parameters
        ----------

        Returns
        -------

        """
        return self.getData()

    @data.setter
    def data(self, newData):
        """ Property of the data of this datawrapper object.

        Parameters
        ----------

        Returns
        -------

        """
        self.setData(newData)

    def setData(self, data):
        """ Replaces the current DataData with specified argument. 
        mh: subclasses can override this to add listener whenevery data is
        replaced.

        Parameters
        ----------

        Returns
        -------

        """
        self._data = data

    def getData(self):
        """ Returns the data in this dw

        Parameters
        ----------

        Returns
        -------
        `self._data` or `None` if '._data' is missing. e.g. subclass overriding `setData` calling `getData` from its `setData` that is called by 'this.__init__'.
        """

        return getattr(self, '_data', None)

    def hasData(self):
        """ Returns whether this data wrapper has data. 


        Parameters
        ----------

        Returns
        -------
        `False` if data does not exist or is `None` or empty, `True` if otherwise.
        """
        d = getattr(self, '_data', None)
        return bool(d)

    def __contains__(self, x):
        """
        """
        try:
            return x in self._data
        except AttributeError:
            return False

    def __len__(self, *args, **kwargs):
        """ 0 if data does not exist or is `None`, size of data if otherwise.

        """

        d = getattr(self, '_data', None)
        return len(d) if d else 0


class DataWrapper(DataContainer, Quantifiable):
    """ A DataWrapper is a composite of data, unit and description.

    mh: note that all data are in the same unit. There is no metadata.
    Implemented from AbstractDataWrapper.
    """

    def __init__(self, *args, **kwds):
        """ 
        """
        super().__init__(*args, **kwds)  # DataWrapper


class DataWrapperMapper():
    """ Object holding a map of data wrappers. """

    def __init__(self, *args, **kwds):  # DataWrapperMapper
        """ 
        """
        super().__init__(*args, **kwds)  # DataWrapperMapper

    def getDataWrappers(self):
        """ Gives the data wrappers, mapped by name.

        Parameters
        ----------

        Returns
        -------

        """
        return self._data



# -*- coding: utf-8 -*-

from fdi.utils.fits_kw import FITS_KEYWORDS, getFitsKw
from fdi.utils.tofits import toFits, fits_dataset
from fdi.dataset.arraydataset import ArrayDataset, Column as aCol
from fdi.dataset.tabledataset import TableDataset
from fdi.dataset.dataset import CompositeDataset
from fdi.dataset.baseproduct import BaseProduct
from fdi.dataset.dateparameter import DateParameter
from fdi.dataset.stringparameter import StringParameter
from fdi.dataset.dataset import Dataset
from fdi.dataset.numericparameter import NumericParameter, BooleanParameter
from fdi.dataset.metadata import MetaData
from fdi.dataset.deserialize import deserialize

try:
    import numpy as np
    from astropy.io import fits
    FITS_INSTALLED = True
except ImportError:
    FITS_INSTALLED = False
    sys.exit(0)

import sys
import os
import pytest

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False


def test_fits_err():
    ima = ArrayDataset(data=[[1, 2, 3, 4], [5, 6, 7, 8]], description='a')
    with pytest.raises(TypeError):
        v = toFits(MetaData())


def test_fits_dataset():
    ima = ArrayDataset(data=[[1, 2, 3, 4], [5, 6, 7, 8]], description='a')
    imb = ArrayDataset(data=[[1, 2, 3, 4], [5, 6, 7, 8], [
                       1, 2, 3, 4], [5, 6, 7, 8]], description='b')
    # im=[[1,2,3,4],[5,6,7,8]]

    hdul = fits.HDUList()
    hdul.append(fits.PrimaryHDU())
    data = [ima, imb]
    u = fits_dataset(hdul, data)
    assert issubclass(u.__class__, fits.HDUList)
    assert len(u) == len(data)+1
    assert list(u[1].data[0]) == [1, 2, 3, 4]
    assert u[1].header[getFitsKw('description')] == 'a'


def test_arr_toFits():
    ima = ArrayDataset(data=[[1, 2, 3, 4], [5, 6, 7, 8]], description='a')
    u = toFits(ima, file=None)
    # when passed a dataset instead of a list, meta go to PrimaryHDU
    assert issubclass(u.__class__, fits.HDUList)
    assert len(u) == 2
    assert list(u[1].data[0]) == [1, 2, 3, 4]
    assert u[0].header[getFitsKw('description')] == 'a'


def test_tab_fits():
    d = {'col1': aCol(data=[1, 4.4, 5.4E3], unit='eV'),
         'col2': aCol(data=[0, 43, 2E3], unit='cnt')}
    d['col1'].type = 'd'
    d['col2'].type = 'i'
    ds = TableDataset(data=d)
    data = [ds]
    # __import__('pdb').set_trace()
    u = toFits(data, file=None)
    assert issubclass(u.__class__, fits.HDUList)
    print('test_tab_fits', u[0])
    assert u[1].columns['col1'].unit == 'eV'
    assert u[1].data[0][0] == str(d['col1'].data[0])
    assert u[1].data[1][1] == str(d['col2'].data[1])


@pytest.fixture(scope='module')
def makecom():
    a1 = [768, 4.4, 5.4E3]
    a2 = 'ev'
    a3 = 'arraydset 1'
    a4 = ArrayDataset(data=a1, unit=a2, description=a3)
    a5, a6, a7 = [[1.09, 289], [3455, 564]], 'count', 'arraydset 2'
    #a8 = ArrayDataset(data=a5, unit=a6, description=a7)
    d = {'col1': aCol(data=[1, 4.4, 5.4E3], unit='eV'),
         'col2': aCol(data=[0, -43, 2E3], unit='cnt')}
    d['col1'].typecode = 'd'
    d['col2'].typecode = 'i'
    a8 = TableDataset(data=d)
    v = CompositeDataset()
    a9 = 'dataset 1'
    a10 = 'dataset 2'
    v.set(a9, a4)
    v.set(a10, a8)
    assert v[a9] == a4
    a11 = 'm1'
    a12 = NumericParameter(description='a different param in metadata',
                           value=2.3, unit='sec')
    v.meta[a11] = a12
    return v


def test_com_fits(makecom):
    v = makecom
    data = [v]
    # __import__('pdb').set_trace()
    u = toFits(data, file=None)
    print(u, len(u))
    assert u[2].data[0] == v['dataset 1'].data[0]
    assert u[2].shape == tuple(v['dataset 1'].shape)
    assert u[3].data['col1'][1] == v['dataset 2']['col1'][1]
    assert u[3].data.dtype[0] == '<f8'


"""
    "A": "B",  # ASCII char
    "L": "B",  # boolean
    "X": "H",  # unsigned short
    "B": "B",  # unsigned char
    "S": "b",  # signed char
    "I": "h",  # signed short
    "U": "H",  # unsigned short
    "J": "i",  # signed integer
    "V": "I",  # unsigned integer
    "K": "l",  # signed long
    "E": "d",  # double
    "D": "d",  # double
    "C": "c",  # complex
    "M": "c"  # complex
"""
tcode = {'b': bool,  # Boolean
         'i8': np.int8,  # 8-bit signed integer
         'i16': np.int16,  # 16-bit signed integer
         'i32': np.int32,  # 32-bit signed integer
         'i64': np.int64,  # 64-bit signed integer
         'u8': np.uint8,  # 8-bit unsigned integer
         'u16': np.uint16,  # 16-bit unsigned integer
         'u32': np.uint32,  # 32-bit unsigned integer
         'u64': np.uint64,  # 64-bit unsigned integer
         'f16': np.float16,  # 16-bit floating point number
         'f32': np.float32,  # 32-bit floating point number
         'f64': np.float64,  # 64-bit floating point number
         'c64': np.complex64,  # 64-bit complex number
         'c128': np.complex128  # 128-bit complex number
         }


def test_toFits_metadata(makecom):
    for ds in [ArrayDataset, TableDataset, CompositeDataset]:
        if issubclass(ds, ArrayDataset):
            ima = ds(data=[[1, 2, 3, 4], [5, 6, 7, 8]], description='a')
        elif issubclass(ds, TableDataset):
            d = {'col1': aCol(data=[1, 4.4, 5.4E3], unit='eV'),
                 'col2': aCol(data=[0, -43, 2E3], unit='cnt')}
            d['col1'].typecode = 'd'
            d['col2'].typecode = 'i'
            ima = ds(data=d)
        elif issubclass(ds, CompositeDataset):
            # __import__('pdb').set_trace()
            ima = makecom
        else:
            assert False
        ima.meta['datetime'] = DateParameter(
            '2023-01-23T12:34:56.789012', description='date keyword')
        ima.meta['quat'] = NumericParameter(
            [0.0, 1.1, 2.4, 3.5], description='q')
        ima.meta['float'] = NumericParameter(
            1.234, description='float keyword')
        ima.meta['integer'] = NumericParameter(
            1234, description='integer keyword')
        ima.meta['string_test'] = StringParameter(
            'abc', description=' string keyword')
        ima.meta['boolean_test'] = BooleanParameter(
            True, description=' boolean keyword')
        imb = ArrayDataset(data=[[1, 2, 3, 4], [5, 6, 7, 8], [
                           1, 2, 3, 4], [5, 6, 7, 8]], description='b')

        data = [ima]
        u = toFits(data, file=None)
        assert issubclass(u.__class__, fits.HDUList)
        #assert len(u) == len(data)+1
        #print("-----", u[0].header)
        w = u[1]
        # print(w.header)
        assert w.header['DATETIME'] == '2023-01-23T12:34:56.789012'
        assert w.header['QUAT0'] == 0.0
        assert w.header['QUAT1'] == 1.1
        assert w.header['QUAT2'] == 2.4
        assert w.header['QUAT3'] == 3.5
        assert w.header['FLOAT'] == 1.234
        assert w.header['INTEGER'] == 1234
        assert w.header['STRING_T'] == 'abc'
        assert w.header['BOOLEAN_'] == 'T'

        if issubclass(ds, ArrayDataset):
            assert w.header['NAXIS1'] == len(ima.data[0])
            assert w.header['NAXIS2'] == len(ima.data)
            assert w.header['NAXIS'] == len(ima.shape)
            assert w.data[0][0] == ima[0][0]
            assert w.data[1][1] == ima[1][1]
        elif issubclass(ds, TableDataset):

            assert w.data[0][0] == d['col1'].data[0]
            assert w.data[1][1] == d['col2'].data[1]
        elif issubclass(ds, CompositeDataset):
            pass


def test_prd_fits(makecom):
    c = makecom
    p = BaseProduct('abc')
    p['com'] = c
    p.meta['creationDate'] = DateParameter(
        '2023-01-23T12:34:56.789012', description='date keyword')
    v = toFits(p, file=None)

    assert v[2].data[0] == c['dataset 1'].data[0]
    assert v[0].header['DESC'] == 'abc'
    assert v[0].header['TYPE'] == 'BaseProduct'
    assert v[0].header['DATE'] == '2023-01-23T12:34:56.789012'
    v = toFits(p, '/tmp/test.fits', overwrite=1)
    f = fits.open('/tmp/test.fits')
    f.close()
    assert f[0].header == v[0].header
    # test stream
    # __import__('pdb').set_trace()
    f = open('/tmp/test.fits', 'wb+')
    stream = toFits(p)
    f.write(stream)
    f.close()
    f = fits.open('/tmp/test.fits')
    f.close()
    assert f[0].header == v[0].header
    del f


def test_Fits_Kw():
    # Fits to Parameter name
    assert FITS_KEYWORDS['DATASUM'] == 'checksumData'
    assert FITS_KEYWORDS['DESC'] == 'description'
    assert FITS_KEYWORDS['CUNIT'] == 'cunit'
    assert getFitsKw('checksumData') == 'DATASUM'
    assert getFitsKw('description') == 'DESC'
    assert getFitsKw('cunit01') == 'CUNIT01'
    assert getFitsKw('description01234') == 'DES01234'
    assert getFitsKw('description01234', ndigits=2) == 'DESC34'
    # with pytest.raises(ValueError):
    assert getFitsKw('checksumData0123', 3) == 'DATAS123'
    assert getFitsKw('checksumData0123', 2) == 'DATASU23'
    with pytest.raises(TypeError):
        # wrong format in `extra`
        assert getFitsKw('foo0123', 5, (('foo', 'BAR'))) == 'BAR00123'
    assert getFitsKw('foo0123', 5, (('foo', 'BAR'),)) == 'BAR00123'




from ..dataset.serializable import serialize
from ..dataset.deserialize import deserialize
from ..utils.getconfig import getConfig
from ..utils.common import trbk, lls
from ..pal import webapi
from .fdi_requests import reqst, cached_json_dumps

from ..httppool.session import requests_retry_session
import aiohttp

import functools
import logging
import sys
import json
import copy

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = str
    from urllib.parse import urlparse
else:
    PY3 = False
    # strset = (str, unicode)
    strset = str
    from urlparse import urlparse

logger = logging.getLogger(__name__)
# logger.debug('level %d' % (logger.getEffectiveLevel()))

session = requests_retry_session()

pcc = getConfig()
defaulturl = 'http://' + pcc['cloud_host'] + \
             ':' + str(pcc['cloud_port'])
default_base = defaulturl + pcc['cloud_api_base'] + \
    '/' + pcc['cloud_api_version']
AUTHUSER = pcc['cloud_username']
AUTHPASS = pcc['cloud_password']


@functools.lru_cache(maxsize=16)
def getAuth(user=AUTHUSER, password=AUTHPASS):
    return HTTPBasicAuth(user, password)


def read_from_cloud(requestName, client=None, asyn=False, server_type='csdb', **kwds):
    """Apply GET method to CSDB server and get reply info back.

    if k-v parameters in kwds are simple values, return simple result,
    or parts of them; is v is a list, return a list of values of each
    member of v.

    Parameters
    ----------
    requestName : str
    client : str, method-func
    asyn : bool
        Run asynchronously. On of the parameters must be a list.
    **kwds :

    Returns
    -------
    obj, list

    Raises
    ------
    ValueError


    """

    if client is None:
        client = session
    header = {'Content-Type': 'application/json;charset=UTF-8'}
    if requestName == 'getToken':
        requestAPI = defaulturl + '/user/auth/token'
        postData = {'username': AUTHUSER, 'password': AUTHPASS}
        res = reqst(client.post, requestAPI, headers=header,
                    data=serialize(postData), server_type=server_type, **kwds)
    elif requestName == 'verifyToken':
        requestAPI = defaulturl + \
            '/user/auth/verify?token=' + kwds.pop('token', '')
        res = reqst(client.get, requestAPI, server_type=server_type, **kwds)
    elif requestName[0:4] == 'info':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        if requestName == 'infoUrn':
            requestAPI = default_base + \
                '/storage/info?urns=' + kwds.pop('urn')
        elif requestName == 'infoPool':
            limit = kwds.pop('limit', 10000)
            getc = 1 if kwds.pop('getCount', 0) else 0
            requestAPI = default_base + \
                f'/storage/info?getCount={getc}&pageIndex=1&pageSize={limit}&pools=' + \
                kwds.pop('pools')
        else:
            raise ValueError("Unknown request API: " + str(requestName))
        res = reqst(client.get, requestAPI, headers=header,
                    server_type=server_type, **kwds)

    elif requestName == 'getMeta':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI = default_base + \
            '/storage/meta?urn=' + kwds.pop('urn')
        res = reqst(client.get, requestAPI, headers=header,
                    server_type=server_type, **kwds)
        return res['_ATTR_meta']
    elif requestName == 'getDataInfo':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI0 = default_base + \
            '/storage/searchByPoolOrPath?limitCount=%d'

        """ result example:
        {
              "code": 0,
              "msg": "OK",
              "data": [
                {
                  "url": "http://...:.../csdb/v1/storage/test_csdb_fdi/fdi.dataset.testproducts.TCC/732",
                  "path": "/test_csdb_fdi/fdi.dataset.testproducts.TCC/732",
                  "urn": "urn:test_csdb_fdi:fdi.dataset.testproducts.TCC:732",
                  "timestamp": 1675267011883,
                  "tags": [],
                  "index": 732,
                  "md5": "2FCC79CA9F0FD0A671D45FAC35528465",
                  "size": 5246,
                  "contentType": null,
                  "fileName": "fdi.dataset.testproducts.TCC",
                  "dataType": "fdi.dataset.testproducts.TCC"
                },
                ...
                      ],
                "total": 84
         }
        """
        # paths can be URNs
        paths = kwds.pop('paths', '')
        pool = kwds.pop('pool', None)

        listpa = isinstance(paths, (list, tuple))
        if not listpa:
            paths = [paths]
        po = f'&pool={pool}' if pool else ''
        # this remembers all members
        pp = []
        # this has only one that has pool
        pp_one_pool = []
        for a in paths:
            if a:
                if a.startswith('urn:'):
                    a = a[3:].replace(':', '/')
                seg = f'&path={a}'
                pp.append(seg)
                pp_one_pool.append(seg)
            else:
                pp.append(po)
                if not pp_one_pool:
                    pp_one_pool.append(po)
        limit = kwds.pop('limit', 10000)
        # max length with one extra
        requestAPI1 = requestAPI0 % (limit)
        if asyn:
            apis = [requestAPI1+p for p in pp_one_pool]
            reses = reqst('get', apis, headers=header,
                          server_type=server_type, **kwds)
            re = dict(zip(pp_one_pool, reses))
        else:
            re = {}
            for p in pp_one_pool:
                requestAPI = requestAPI1 + p
                r = reqst(client.get, requestAPI, headers=header,
                          server_type=server_type, **kwds)

                re[p] = r
        # reconstruct
        res = [re[x] for x in pp]
        return res if listpa else res[0]

    elif requestName == 'getDataType':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')

        requestAPI = default_base + \
            '/datatype/list'
        res = reqst(client.get, requestAPI, headers=header,
                    server_type=server_type, **kwds)

    elif requestName == 'uploadDataType':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        header["accept"] = "*/*"
        # somehow application/json will cause error "unsupported"
        del header['Content-Type']  # = 'application/json'  # ;charset=UTF-8'
        requestAPI = default_base + \
            '/datatype/upload'
        cls_full_name = kwds.pop('cls_full_name')
        jsn = cached_json_dumps(cls_full_name,
                                ensure_ascii=kwds.get('ensure_ascii', True),
                                indent=kwds.get('indent', 2))
        fdata = {"file": (cls_full_name, jsn)}
        data = {"metaPath": "/metadata",
                "dataType": cls_full_name}
        res = reqst(client.post, requestAPI,
                    files=fdata, data=data, headers=header, server_type=server_type, **kwds)
    elif requestName == 'delDataTypeData':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI = default_base + \
            f'/storage/delDatatypeData?path=' + kwds.pop('path')
        res = reqst(client.delete, requestAPI, headers=header,
                    server_type=server_type, **kwds)
    elif requestName == 'delDataType':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI0 = default_base + \
            f'/storage/delDatatype?path='
        _p = kwds.pop('path')
        if isinstance(_p, str):
            paths = [_p]
            alist = False
        else:
            paths = _p
            alist = True
        apis = [requestAPI0+p for p in paths]
        if asyn:
            res = reqst('delete', apis, headers=header,
                        server_type=server_type, **kwds)
        else:
            rs = []
            for a in apis:
                r = reqst(client.delete, a, headers=header,
                          server_type=server_type, **kwds)
                rs.append(r)
            res = rs if alist else rs[0]
    elif requestName == 'remove':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI0 = default_base + \
            '/storage/delete?path='
        _p = kwds.pop('path', '')
        if isinstance(_p, str):
            paths = [_p]
            alist = False
        else:
            paths = _p
            alist = True

        apis = [requestAPI0+p for p in paths]
        if asyn:
            r = reqst('post', apis, headers=header,
                      server_type=server_type, **kwds)
            rs = [0 if x is None else 1 for x in r]
        else:
            rs = []
            for a in apis:
                r = reqst(client.post, a, headers=header,
                          server_type=server_type, **kwds)
                rs.append(0 if r is None else 1)
        res = rs if alist else rs[0]
    elif requestName == 'existPool':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI = default_base + \
            '/pool/info?storagePoolName=' + kwds.pop('poolname')
        res = reqst(client.get, requestAPI, headers=header,
                    server_type=server_type, **kwds)
    elif requestName == 'createPool':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI = default_base + \
            '/pool/create?poolName=' + kwds.pop('poolname') + '&read=0&write=0'
        res = reqst(client.post, requestAPI, headers=header,
                    server_type=server_type, **kwds)
    elif requestName == 'wipePool':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI = default_base + \
            '/pool/delete?storagePoolName=' + kwds.pop('poolname')
        res = reqst(client.post, requestAPI, headers=header,
                    server_type=server_type, **kwds)
    elif requestName == 'restorePool':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI = default_base + \
            '/pool/restore?storagePoolName=' + kwds.pop('poolname')
        res = reqst(client.post, requestAPI, headers=header,
                    server_type=server_type, **kwds)
    elif requestName == 'addTag':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI0 = default_base + \
            '/storage/addTags?tags='

        _t = kwds.pop('tag')
        if isinstance(_t, str):
            tags = [_t]
            alist = False
        else:
            tags = _t
            alist = True
        u = '&urn=' + kwds.pop('urn')
        apis = [requestAPI0+t+u for t in tags]
        if asyn:
            res = reqst('get', apis, headers=header,
                        server_type=server_type, **kwds)
        else:
            rs = []
            for a in apis:
                r = reqst(client.get, a, headers=header,
                          server_type=server_type, **kwds)
                rs.append(r)
            res = rs if alist else rs[0]
    elif requestName == 'tagExist':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI = default_base + \
            '/storage/tagExist?tag=' + kwds.pop('tag')
        res = reqst(client.get, requestAPI, headers=header,
                    server_type=server_type, **kwds)
    elif requestName == 'getUrn':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI = default_base + \
            '/storage/tag?tag=' + kwds.pop('tag')
        res = reqst(client.get, requestAPI, headers=header,
                    server_type=server_type, **kwds)
    else:
        raise ValueError("Unknown request API: " + str(requestName))
    # print("Read from API: " + requestAPI)
    # must remove csdb layer
    return res


def _multi_input_header(kwds, n, header):

    _k = kwds.pop('token', '')
    withtokens = []
    for tok in (_k if isinstance(_k, list) else ([_k] * n)):
        w = copy.copy(header)
        w['X-AUTH-TOKEN'] = tok
        withtokens.append(w)

    _h = kwds.pop('header', {})
    if not isinstance(_h, list):
        for _ in withtokens:
            _.update(_h)
        headers = withtokens
    else:
        for hdr, tok in zip(_h, withtokens):
            hdr.update(tok)
        headers = _h
    return headers


def load_from_cloud(requestName, client=None, asyn=False, server_type='csdb', **kwds):
    if client is None:
        client = session
    header = {'Content-Type': 'application/json;charset=UTF-8'}
    requestAPI = default_base

    if requestName == 'uploadProduct':
        # application/json causes "only allow use multipart/form-data"
        del header['Content-Type']
        header['X-CSDB-AUTOINDEX'] = '1'
        header['X-CSDB-METADATA'] = '/_ATTR_meta'
        header['X-CSDB-HASHCOMPARE'] = '0'

        requestAPI0 = requestAPI + \
            '/storage/upload'
        _p = kwds.pop('path', '')
        if isinstance(_p, str):
            paths = [_p]
            alist = False
        else:
            paths = _p
            alist = True
        apis = [
            (f'{requestAPI0}?path={p}' if p else requestAPI0) for p in paths]
        # all parameters, if is given a single value, will be expanded to a list of this size.
        n = len(apis)

        _pr = kwds.pop('products', None)
        if not isinstance(_pr, list):
            # XXX TODO: a better way to determine seriaized product
            prds = [_pr] * n
        else:
            prds = _pr

        _f = kwds.pop('resourcetype')
        if isinstance(_f, str):
            fileNames = [_f] * n
        else:
            fileNames = _f

        _t = kwds.pop('tags', None)
        if isinstance(_t, str) or _t is None:
            tags = [_t] * n
        else:
            tags = _t

        if asyn:
            # data = [{'file': (f, p), 'tags': t}
            #         for f, p, t in zip(fileNames, prds, tags)]
            data = []
            for f, p, t in zip(fileNames, prds, tags):
                d = aiohttp.FormData()
                d.add_field('file', p,
                            content_type='application/json', filename=f)
                if t:
                    d.add_field('tags', t)
                data.append(d)
        else:
            files = [{'file': (f, p)} for f, p in zip(fileNames, prds)]
            data = [{'tags': t} for t in tags]

        headers = _multi_input_header(kwds, n, header)

        serialize_out = kwds.pop('serialize_out', '')
        if asyn:
            res = reqst('post', apis, data=data,
                        headers=headers, server_type=server_type, **kwds)
        else:
            res = []
            for a, f, d, h in zip(apis, files, data, headers):
                r = reqst(client.post, a, files=f, data=d,
                          headers=h, server_type=server_type, **kwds)
                res.append(r)
        return res if alist else res[0]

    elif requestName == 'pullProduct':
        # header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        # requestAPI = requestAPI + '/storage/get?urn=' + kwds.pop('urn', '')
        # res = reqst(client.get, requestAPI,
        #             headers=header, stream=True, server_type=server_type, **kwds)
        # TODO: save product to local

        requestAPI0 = requestAPI + '/storage/get?urn='
        _u = kwds.pop('urn', '')
        if isinstance(_u, list):
            alist = True
        else:
            urns = [_u]
            alist = False

        n = len(urns)

        headers = _multi_input_header(kwds, n, header)
        apis = [requestAPI0 + u for u in urns]

        if asyn:
            res = reqst('get', apis, headers=headers,
                        server_type=server_type, **kwds)
        else:
            res = []
            for a, h in zip(apis, headers):
                r = reqst(client.get, a, headers=h,
                          stream=True, server_type=server_type, **kwds)
                res.append(r)
        return res if alist else res[0]
    else:
        raise ValueError(f'Unknown request API: {requestName}')


def delete_from_server(requestName, client=None, asyn=False, server_type='csdb', **kwds):
    if client is None:
        client = session
    header = {'Content-Type': 'application/json;charset=UTF-8'}
    requestAPI = default_base
    if requestName == 'delTag':
        header['X-AUTH-TOKEN'] = kwds.pop('token', '')
        requestAPI0 = requestAPI + '/storage/delTag?tag='

        _t = kwds.pop('tag', None)
        if isinstance(_t, str) or _t is None:
            tags = [_t]
            alist = False
        else:
            tags = _t
            alist = True
        apis = [requestAPI0 + t for t in tags]
        if asyn:
            res = reqst('delete', apis,
                        headers=header, server_type=server_type, **kwds)
        else:
            rs = []
            for a in apis:
                r = reqst(client.delete, a, headers=header,
                          server_type=server_type, **kwds)
                rs.append(r)
            res = rs if alist else rs[0]
    else:
        raise ValueError("Unknown request API: " + str(requestName))
    # print("Read from API: " + requestAPI)
    return res


def get_service_method(method):
    service = method.split('_')[0]
    serviceName = method.split('_')[1]
    if service not in webapi.PublicServices:
        return 'home', None
    return service, serviceName



# -*- coding: utf-8 -*-


from collections import ChainMap
import sys
import logging
import copy
from functools import lru_cache

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

Load_Failed = object()
""" unique object to mark this failure condition."""


def refloader(key, mapping, remove=True, exclude=None, ignore_error=False):
    """ Generates key-value pair out of a map containing name-content
    pairs, by referencing.

    Subclasses should override this function unless this name space
    contains the same kind of items in `default`

    Parameters
    ----------
    key: str
        name in the key-value pair.
    mapping : dict
        a map containing name-content pairs (such as
    `default`, `initial`).
    remove : bool
        if set, remove this pair from the source after loading.
    exclude : list
        A list of keys to avoid loading. Default None.
    ignore_error : boolean
        Do not throw exception when error happens during loading. Log, set `Load_Failed`, and Move on.

    Returns
    -------
   dict:
       key and load-result pairs. load-result is `Load_Failed` if loading of the key was not successful.
    """

    if key in exclude:
        res = Load_Failed
    else:
        res = mapping.get(key, Load_Failed)
    if remove and res is not Load_Failed:
        del mapping[key]
    # return key in the mapping and the load result.
    return {key: res}


class NameSpace_meta(type):
    """ metaclass for name-spaces such as class white list and schemas.

    Ref 'classproperty'.   # https://stackoverflow.com/a/1800999
    """

    sources = [{}]
    """ name-content list from the main package and for plug-in/app."""

    def __new__(metacls,  clsname, bases, attrs,
                sources=None,
                extensions=None,
                load=None, **kwds):
        """ Internal map is initialized with `sources`.

            The internal map is initialized with a `default`
        and a list of `extension` maps which can be
        collection of key-value pairs. These maps are put
        into the `sources` map. However these maps are only
        the information needed to populate the main map, the
        target map of namespace.

        The target namespace is also represented by a
        collection of key-value pairs but each of them
        reside in a cache map, and are loaded into the cache
        map by the `load` function lazily when the key is
        used. The default `refloader` just copies the
        reference of the values in the `sources` map by the
        same name.

        When looking up a key, the cache maps will be
        searched left to right.

        This architecture allows expensive values to be
        associated with names gradually in a cache in a
        pay-as-you-need manner.

        Examples
        --------

        For an app package with many classes:

        Import user classes in a python file for example
        projectclasses.py:

        .. code-bloc::

        clz_map = {
                'MyClass1': 'mypackage.mymodule',
                'MyClass2': 'mypackage.mymodule'
        }

        # from another module defining a dict of
        # Class_name: Class_obj pairs

        try:
            from mypackage.mymodule import pairs
        except (ImportError, ModuleNotFoundError) as e:
            logger.info(e)
            raise

        from fdi.dataset.namespace import NameSpace_meta

        def loader():
            ...

        class PC(metaclass=NameSpace_meta,
                  sources=[Reverse_Modules_Classes, pairs, clz_map],
                  load=loader
                  ):
              pass

        prjcls = PC.mapping

        new_instance = prjcls['MyClass1']

        Define new classes and update `PC`::

        class Myclass():
              ...

        PC.update({'foo': MyClass})

        and use::

        ``new_instance = PC.mapping['foo']``

        Parameters
        ----------
        clsname: str
            Name of the class being defined(Event in this example)
        bases: tuple
            Base classes of the constructed class, empty tuple in this case
        attrs: dict
            Dict containing methods and fields defined in the class
        sources: list
            A list of maps containing the core/platform/framework/primary package namespace and plug-in/application package name spaces.
        extensions: list
            A list of key-value maps to extend the `cache`.
        load: function
            classmethod to load a key from `initial` of the internal map.
        kwds: dict
            member `key`-`val` pairs: `k` will be added to instance-classes' class attributes namespace, initiated to `val`

        Returns
        -------
        cls
            new class
        """
        new_cls = super().__new__(metacls, clsname, bases, attrs)

        if sources is None:
            sources = metacls.sources
        if load is None:
            # defined in this module
            load = refloader

        nm = Lazy_Loading_ChainMap(*sources, extensions=extensions, load=load)
        if kwds:
            for name, value in kwds.items():
                setattr(new_cls, name, value)
        logger.debug('***maps*** %s' % str(sources)[:300])
        new_cls._the_map = nm
        new_cls.mapping = nm

        logger.debug("New class made with metaclass %s: _the_map 0x%x, sources %d, initial %d, cache %d. load %s, kwds %s,initial=%s..." %
                     (metacls.__name__,
                      id(nm),
                      len(nm.sources),
                      len(nm.initial),
                      len(nm.cache),
                      str(load)[:300],
                      str(kwds)[:300],
                      str(nm.initial)[:300]
                      ))
        return new_cls

    def clear(cls):
        """ Empty the internal mapping including `maps[1:]`.

        `sources` map is not wiped.

        """
        for m in cls._the_map.maps:
            m.clear()

    def update(cls, *args, **kwds):
        """ Updates the mapping.

        Parameters
        ----------
        c: Mapping to be used to update the main map with. Subclasses that need to
        load must format key and values as required.

        Returns
        -------
        dict: The mapping.

        """
        cls._the_map.update(*args, **kwds)
        return cls._the_map

    def reload(cls):
        """ re-import classes in the map.

        """

        cls._the_map.reload()
        return cls._the_map


class Lazy_Loading_ChainMap(ChainMap):
    """ A mapping the populates its main storage as needed
    from source and extension maps.

    Implementated with a `ChainMap` of a cache, a initial,
    and an arbitrary number of extension maps.

    The name (the key) is searched in the cache and, if not
    found, in other maps on the chain. the `load` function is
    used to do the loading.
    """

    def __init__(self, *args, extensions=None, load=None, **kwds):

        if extensions is None:
            extensions = []
        if load is None:
            load = refloader
        self.load = load

        self.extenions = extensions

        # failed = {}
        # """ name-content pairs of the unloadable pairs from `default`. """

        self.cache = dict()
        """ for the loaded key-vals. """

        self.exclude = list()
        """exclude : list
            A list of keys to avoid loading. Default None."""

        self.ignore_error = False
        """ignore_error : boolean
            Do not throw exception when error happens during loading. Log, set `Load_Failed`, and Move on."""

        self.sources = ChainMap(*args, **kwds)

        self.initial = dict(self.sources)
        """ This mapping stores name-content pairs that are used to
        build the main map (the `cache`). It is dict-wrapped `sources`.
        Example: module_name-classe_names, schema
        store, configs."""

        super().__init__(self.cache, self.initial, *extensions)

        logger.debug("New LLC %s initialized: _the_map 0x%x sources %d, initial %d, cache %d. extensions %d. initial=%s..." %
                     (self.__class__.__name__,
                      id(self),
                         len(self.sources),
                         len(self.initial),
                         len(self.cache),
                         len(self.extenions),
                         str(self.initial)[:300]
                      ))

    def __getitem__(self, key):

        for i, m in enumerate(self.maps):
            if m is self.initial:
                loaded = self.load(key, self.initial, remove=True,
                                   exclude=self.exclude, ignore_error=self.ignore_error)
                for k, re in loaded.items():
                    if re is not Load_Failed:
                        # success. put into cache
                        self.cache[k] = re
                        # is what we look for
                        if k == key:
                            return re
                    else:
                        #  ignore to let future calls try.
                        pass
                # reaching this point means not found in loaded.
                continue
            else:
                # is it excluded?
                if key in m:
                    if key in self.exclude:
                        logger.debug(
                            'Find "%s" in map%d but it is excluded.' % (key, i))
                        return None
                    return m[key]
        if not self.ignore_error:
            raise KeyError('Key "%s" not found in any namespace.' % key)
        return None

    def __setitem__(self, key, value):

        if key in self.initial:
            self.initial.__delitem__(key)
        self.cache.__setitem__(key, value)
        return

    def add_ns(self, ns=None, order=0):
        """ Add new name space in the list of internal ones.

        Parameters
        ----------
        order: int
            The number of maps to look up before this one
        is . If negative, `-n` means the n-th from the last.
        E.g. `order=-1` means to become the last one.
        ns: mapping
            Namespace map to be looked up.
        """
        if ns is None:
            ns = {}
        if order == -1:
            self.maps.append(ns)
        elif order < -1:
            self.maps.insert(order+1, ns)
        else:
            self.maps.insert(order, ns)
        return self

    def update(self, c=None, exclude=None, verbose=False,
               extension=None, ignore_error=False,
               ):
        """ Updates the mapping.

        Parameters
        ----------
        c: mapping
            to be used to update with. Subclasses that need to
        load must format key and values as required.
        exclude: boolean
            Ignore these keys when updating.
        extension: mapping
            add `c` as a new sources map.

        Returns
        -------
        dict: The mapping.

        """
        if exclude is None:
            exclude = []
        cc = copy.copy(c)
        if cc:
            for x in exclude:
                cc.pop(x, None)
            if extension:
                self.sources.maps.insert(0, cc)
            else:
                ini = self.initial
                in_initial = set(cc.keys()) & set(ini.keys())
                for i in in_initial:
                    ini.__delitem__(i)
                self.cache.update(cc)

        return self

    def reload(self):
        """ Update the `initial` map with `sources`, empty other maps.

        Parameters
        ----------

        Returns
        -------
        ChainMap:
            `self`.
        """

        self.clear()
        self.initial.update(self.sources)
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ..dataset.namespace import NameSpace_meta, Load_Failed
from ..dataset.readonlydict import ReadOnlyDict
from ..pns.config import pnsconfig as builtin_conf
from requests.auth import HTTPBasicAuth
from os.path import join, expanduser, expandvars, isdir
import functools
import socket
import getpass
import json
import os
import argparse
import copy
import sys
import importlib

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('logging level %d' % (logger.getEffectiveLevel()))

Config_NameSpace = {}
""" `NameSpace` managers for configuration key-value pairs."""

Config_Look_Up = {}
""" `Lazy_Loading_ChainMap`s for getting value with a key from Configuration dictionaries in files and in OS Environment.. """


@functools.lru_cache(8)
def get_file_conf(conf_name):
    """ figure iut config file name and returns the contents.

    :conf_name: str, 'pns' etc.

    Returns: dict for pre-fixed variables and for not prefixed variables.
    """

    CU = conf_name.upper() + '_'
    envname = CU + 'CONF_DIR'
    epath = os.getenv(envname, '')
    if isdir(epath):
        confp = epath
    else:
        # environment variable <conf_name>_CONFIG_DIR is not set
        env = expanduser(expandvars('$HOME'))
        # apache wsgi will return '$HOME' with no expansion
        if env == '$HOME':
            env = '/root'
        confp = join(env, '.config')
    # this is the var_name part of filename and the name of the returned dict
    var_name = conf_name+'config'
    module_name = conf_name+'local'
    file_name = module_name + '.py'
    filep = join(confp, file_name)
    absolute_name = importlib.util.resolve_name(module_name, None)
    logger.debug('Configuration file %s/%s. absolute mod name %s' %
                 (confp, file_name, absolute_name))
    # if sys.path[0] != confp:
    #    sys.path.insert(0, confp)
    # print(sys.path)
    # for finder in sys.meta_path:
    #     spec = finder.find_spec(absolute_name, filep)
    #     print(spec)  # if spec is not None:

    try:
        # print('zz', spec)
        module = sys.modules.get(module_name, None)

        if 0 and module:
            # module has been imported. clear cache and re-read
            # importlib.invalidate_caches()
            module = importlib.reload(module)
            # c = getattr(nm, var_name)
            logger.debug(f'Module {module_name} to be reloaded.')
        spec = importlib.util.spec_from_file_location(absolute_name, filep)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        logger.debug('Loaded %s/%s.' % (confp, file_name))
        sys.modules[module_name] = module
        # the following suffers from non-updating loader
        # importlib.invalidate_caches()
        # module = importlib.import_module(module_name)
        # modul = __import__(module_name, globals(), locals(), [var_name], 0)

    except (ModuleNotFoundError, FileNotFoundError) as e:
        logger.warning(str(
            e) + '. Use default config in the package in fdi/pns/config.py. Copy it to ~/.config/%slocal.py and make persistent customization there.' % conf_name)
        return {}, {}
    # return a copy so refs can be wiped out by deleting the dict variable.
    return copy.deepcopy(getattr(module, var_name, {})), copy.deepcopy(getattr(module, 'config', {}))


def getMappings(conf_name, builtin):
    """return dictionaries of variable name in Environment or configure file.

    The look-up order is:

        1. with prefix in the Environment, e.g. `PNS_HOST`.
        2. in `prefixed_config`, e.g. `prefixed_config['host']`.
        3. without prefix in the Environment, e.g. `HOST`.
        4. in `config`, e.g. `config['host']`.

    Parameters
    ----------
    prefixed_config : dict
        configuration map of variables with prefix in Environment.
    config : dict
        configuration map of variables without prefix in Environment.
    conf_name : str
        The prefix. e.g. 'pns'.
    osenviron : dict
        Environment variables.
    name : str
        Variable name.

    Returns
    -------
    str
        Variable value.

    Exception
    ---------
    LookupError
        Variable not found.
    """

    # 'pns' etc
    skip = len(conf_name) + 1
    cnu = conf_name.upper() + '_'
    prefixed, not_prefixed = get_file_conf(conf_name)

    if 0:
        prefixed_osenv = dict((k[skip:].lower(), v)
                              for k, v in os.environ.items() if k[:skip] == cnu and k[skip:].lower() in prefixed)
    else:
        # prefixed names in ENV but not in config files allowed
        prefixed_osenv = dict((k[skip:].lower(), v)
                              for k, v in os.environ.items() if k[:skip] == cnu)

    not_prefixed_osenv = dict((k, os.environ[k.upper()])
                              for k, v in not_prefixed.items() if k.upper() in os.environ)

    if builtin is None:
        builtin = {}

    # higher proority is on the left
    return prefixed_osenv, prefixed, not_prefixed_osenv, not_prefixed, builtin


def loader(key, mapping, remove=True, exclude=None, ignore_error=False):

    if key in exclude:
        res = Load_Failed
    else:
        res = mapping.get(key, Load_Failed)
    if remove and res is not Load_Failed:
        del mapping[key]
    # return key in the mapping and the load result.
    return {key: res}


@functools.lru_cache(8)
def getConfClass(conf_name, builtin):
    if issubclass(builtin.__class__, ReadOnlyDict):
        builtin = dict(builtin)
    maps = getMappings(conf_name, builtin)

    class Conf(metaclass=NameSpace_meta,
               sources=maps,
               load=loader
               ):
        pass
    return Conf


_url_mark = 'poolurl:'
_len_um = len(_url_mark)


def cget(name, conf_name='pns', builtin=None, force=False):
    """Find value for configured variables.

    Look up in various places and check if `poolurl` is needed.
    The look-up order is:

        1. with prefix in the Environment, e.g. `PNS_HOST`.
        2. in `prefixed_config`, e.g. `prefixed_config['host']`.
        3. without prefix in the Environment, e.g. `HOST`.
        4. in `config`, e.g. `config['host']`.


    Parameters
    ----------
    name : str
        variable name. If starting with 'poolurl:' the rest of the
        name is used as the poolname part of a default-spec poolurl.
    conf_name : str
        name of configuration. Default 'pns'.
    builtin : dict
        a built-in dict given as the basis.
    force : bool
        Re-make the `Config_NameSpace` entry named `conf_name`, no matter what other parameters are.

    Returns
    -------
    str
        value of the variable.

    """

    global Config_NameSpace
    global Config_Look_Up

    if builtin is None:
        builtin = {}

    if force:
        getConfClass.cache_clear()

    conf_class = getConfClass(conf_name, ReadOnlyDict(builtin))
    Config_NameSpace[conf_name] = conf_class
    clu = conf_class.mapping
    Config_Look_Up[conf_name] = clu

    if not name:
        # name not given
        logger.debug(f'Dumping config {conf_name}.')
        return clu

    # check if request poolurn
    for_poolurl = name.startswith(_url_mark)

    if for_poolurl:
        name = name[_len_um:]
        # return poolurl if name startswith `poolurl`
        logger.debug(f'Getting poolurl by {name}.')
        purl = ''.join((clu['scheme'], '://',
                        clu['host'], ':',
                        str(clu['port']),
                        clu['baseurl']
                        ))
        # with the name
        return '%s/%s' % (purl, name)

    var = conf_class.mapping[name]
    logger.debug(f'Got config for {name} : {var}.')

    return var


def getConfig(name=None, conf='pns', builtin=builtin_conf, force=False):
    """Imports a dict named [conf]config.

    The contents of configuration are the key-value pairs of a
    `dict` variable :mod:`fdi.pns.config::<conf>config` by default.

    The configuration is updated by contents of a
    configuration file in the same format as `fdi.pns.config:pnsconfig`.
    Name of the configuration file is in the form of `<conf>local.py`
    where `<conf>` is the value of the `conf` parameter of this
    function, 'pns' by default.

    The config file directory is the process owner's ``~/.config/``
    by default. It can be modified by the environment
    variable ``<uppercased conf>_CONF_DIR``, e.g. `PNS_CONF_DIR`..

    An exisiting configuration value can be overridden
    by that of an environment variable. The env var is named
    `<uppercased <name>` for name variables in the `config` dict;
    but named `<uppercased <conf>_<name>` for name variables in
    the pre-fixed dict.
    For example configuration of `host` in `pnsconfig` dict is
    overridden by the value of envirionment variable
    `PNS_HOST`. `kc_type` in `config` dict is overridden by `KC_TYPE`.

    Parameters
    ----------
    name : str
        Identifier of the configured item whose value will be returned.
        If started with ```poolurl:```, construct a poolurl with
        ```scheme``` and ```node``` with ```/{name}``` at the end.
        Default ```None```, a mapping of all configured items
        corrected with envirionment variables is returned. 
        Prefixed names in ENV but not in config files are allowed.
    conf : str
         File `<conf>local.py`` defines configuration key-value
         pairs in `dict` named `<conf>config. Default 'pns', so the
         file is 'pnslocal.py', and the variable is `pnsconfig`.
    builtin : dict. To be updated by `<conf>local`. default is `fdi.pns.config`.
    force : bool
        Always 'False'
        reload from file instead of cache for all `conf`s cached.

    Returns
    -------
    obj
        configured value.

    """

    # default configuration is provided. Copy pns/config.py to ~/.config/pnslocal.py
    conflc = conf+'local'
    # this will cause 'spec not found
    # if force and conflc in sys.modules:
    #     logger.debug('Clearing config caches.')
    #     get_file_conf.cache_clear()

    if name:
        name = name.strip()

    return cget(name, conf_name=conf, builtin=builtin, force=force)


# Init Config_NameSpace and Config_Look_Up

cget('')


def make_pool(pool, conf='pns', auth=None, wipe=False):
    """ Return a ProductStorage with given pool name or poolURL.

    :name: PoolURL, or pool name (has no "://"), in which case a pool URL is made based on the result of `getConfig(name=pool, conf=conf)`. Default is ''.
    :auth: if is None will be set to `HTTPBasicAuth` using the `config`.
    :conf: passed to `getconfig` to determine which configuration. Default ```pns```.
    :wipe: whether to delete everything in the pool first.

    Exception
    ConnectionError
    """

    pc = getConfig()
    if '://' in pool:
        poolurl = pool
    else:
        poolurl = getConfig(_url_mark+pool)

    if auth is None:
        auth = HTTPBasicAuth(pc['username'], pc['password'])
    logger.info("PoolURL: " + poolurl)

    # create a product store
    from ..pal.productstorage import ProductStorage
    pstore = ProductStorage(poolurl=poolurl, auth=auth)
    if wipe:
        logger.info('Wiping %s...' % str(pstore))
        pstore.wipePool()
        # pstore.getPool(pstore.getPools()[0]).removeAll()
    # see what is in it.
    # print(pstore)

    return pstore


def get_mqtt_config():
    """ Get configured MQTT info from project configuration file.

    Overrideable by uppercased environment variables.
    Note that there is a 'PNS_' in the beginning environment variable
    name, e.g. ```PNS_MQ_HOST``` for ```pc['mq_host']```
    ref `fdi.utils.getConfig` and your local ```~/.config/pnslocal.py```
    """
    pc = getConfig()
    # default mqtt settings
    mqttargs = dict(
        mq_host=pc['mq_host'],
        mq_port=int(pc['mq_port']),
        mq_user=pc['mq_user'],
        mq_pass=pc['mq_pass'],
        qos=1,
        clean_session=True,
        client_id=socket.gethostname()+'_' + getpass.getuser()+'_' + str(os.getpid())
    )
    return mqttargs


if __name__ == '__main__':

    logger = logging.getLogger()
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("name_pos", metavar='NAME', nargs='?',
                        help="Value of the name parameter in the config file.")
    parser.add_argument("-n", "--name",
                        default=None, help="Value of the name parameter in the config file.")
    parser.add_argument("-c", "--conf",
                        default='pns', help="Configuration ID. default 'pns', so the file is 'pnslocal.py'.")
    parser.add_argument("-f", "--force",  action='store_true',
                        default=False, help="")
    parser.add_argument("-d", "--debug",  action='store_true',
                        default=False, help="")

    args, remainings = parser.parse_known_args(args=sys.argv[1:])

    if args.debug:
        print(f'args: {args}')
    name0 = args.name_pos if args.name is None else args.name
    conf = getConfig(name0, conf=args.conf, force=args.force)
    if issubclass(conf.__class__, dict):
        # dictionart of all config items.
        print(json.dumps(conf, indent=4))
    else:
        print(conf)
    sys.exit(0)

#########


def XXXcheck_env(prefixed_config, config, conf_name, osenviron, name):
    """look up variable name in Environment or configure file.

    Parameters
    ----------
    prefixed_config : dict
        configuration map of variables with prefix in Environment.
    config : dict
        configuration map of variables without prefix in Environment.
    conf_name : str
        The prefix. e.g. 'pns'.
    osenviron : dict
        Environment variables.
    name : str
        Variable name.

    Returns
    -------
    str
        Variable value.

    Exception
    ---------
    LookupError
        Variable not found.
    """

    pcn = '%s_%s' % (conf_name, name)
    for cn, conf in [(pcn, prefixed_config), (name, config)]:
        env_var = cn.upper()
        if env_var in osenviron:
            logger.debug(f'found value for {name} in Env.')
            return osenviron[env_var]
        else:
            if name in conf:
                return conf[name]
            else:
                pass
    raise LookupError(f'{name} not found in config or Environment')


def XXXcget(name, conf_name='pns', builtin=None, force=False):
    """Find value for configured variables.

    Look up in various places and check if `poolurl` is needed. For
    order of places to look-up see :func:`check_env`.

    Parameters
    ----------
    name : str
        variable name. If starting with 'poolurl:' the rest of the
        name is used as the poolname part of a default-spec poolurl.
    conf_name : str
        name of configuration. Default 'pns'.
    builtin : dict
        a built-in dict given as the basis.
    force : bool
        updating source maps, no matter what other parameters are.

    Returns
    -------
    str
        value of the variable.

    """

    osenviron = os.environ

    if builtin is None:
        builtin = {}

    prefixed, not_prefixed = get_file_conf(conf_name)
    config = copy.copy(builtin)
    config.update(not_prefixed)

    if name is None:
        # name not given
        logger.debug(f'Dumping config {conf_name}.')
        res = config
        for k, v in prefixed.items():
            cn = '%s_%s' % (conf_name, k)
            env_var = cn.upper()
            res[k] = osenviron.get(env_var, v)
        return res
    withEnv = functools.partial(check_env, prefixed, not_prefixed,
                                conf_name, osenviron)
    # check if request poolurn
    for_poolurl = name.startswith(_url_mark)

    if for_poolurl:
        name = name[_len_um:]
        # return poolurl if name startswith `poolurl`
        logger.debug(f'Getting poolurl by {name}.')
        purl = ''.join((withEnv('scheme'), '://',
                        withEnv('host'), ':',
                        str(withEnv('port')),
                        withEnv('baseurl')
                        ))
        # with the name
        return '%s/%s' % (purl, name)

    # check env first, then prefixed, then not prefixed env, then not prefixed
    var = withEnv(name)

    return var



# -*- coding: utf-8 -*-

import jsonpath_ng.ext as jex

from functools import lru_cache
import itertools
from collections import UserDict


JEXP = jex.parse('$..*')


def flatten_compact(roots, num=False, style='short', sep='.', path_list=False):
    """ Colapse datapaths of elements of an object.

    Duplicated paths will only have the last one. For example::

    for n, c in {'group1.val': a10['col1'], 'group1.err': a10['col2'],
                 'no-group.val': a10['col1'],
                 'group2.val': a10['col1'], 'group2.err': a10['col2'],
                 'group2.seq': a10['col2'],
                 'group2.wgt': Column(data=['', 0.32, -9876543210], unit='g'),
                 }.items():
        v.addColumn(n, c)

will make one-level datapath::

    col1
    col2
    group1.val
    group1.err
    no-group.val
    group2.val
    group2.err
    group2.seq
    group2.wgt

shown in a tree::

    |__ 'col1'                                     <Column> (3,)
    |__ 'col2'                                     <Column> (3,)
    |__ 'group1.val'                               <Column> (3,)
    |__ 'group1.err'                               <Column> (3,)
    |__ 'no-group.val'                             <Column> (3,)
    |__ 'group2.val'                               <Column> (3,)
    |__ 'group2.err'                               <Column> (3,)
    |__ 'group2.seq'                               <Column> (3,)
    \__ 'group2.wgt'                               <Column> (3,)


Flatten_compact  ```print('\n'self.join(flatten_compact([v]).keys()))```::

    col1
    col2
    g.val
    g.err
    n.val
    g.seq
    g.wgt

The 'group' part  look good in a table header::

    +-----------------+-----------------+--------+----------------------------------------+
    |                 |     group1      |  no-   |                 group2                 |
    |                 |                 | group  |                                        |
    +--------+--------+--------+--------+--------+--------+--------+--------+-------------+
    |   col1 |   col2 |    val |    err |    val |    val |    err |    seq | wgt         |
    |   (eV) |  (cnt) |   (eV) |  (cnt) |   (eV) |   (eV) |  (cnt) |  (cnt) | (g)         |
    +========+========+========+========+========+========+========+========+=============+
    |    1   |    0   |    1   |    0   |    1   |    1   |    0   |    0   |             |
    +--------+--------+--------+--------+--------+--------+--------+--------+-------------+
    |    4.4 |   43.2 |    4.4 |   43.2 |    4.4 |    4.4 |   43.2 |   43.2 | 0.32        |
    +--------+--------+--------+--------+--------+--------+--------+--------+-------------+
    | 5400   | 2000   | 5400   | 2000   | 5400   | 5400   | 2000   | 2000   | -9876543210 |
    +--------+--------+--------+--------+--------+--------+--------+--------+-------------+


    :roots: where. A list of maps.
    :num: preceed keys with sequence numbers.
    :style: for keys, `short`: use shortened path e.g. ```abc.def.hgi``` ```a.d.hgi```. `last2`: use the right-most 2 segments, `full` for untrated paths, anything else to use only the last one.
    :sep: separater used in output. Default is '.'.
    :path_list: a list of path segments in place of value. Defalut `False`.
    :return: dict(flatten_compact_path:(list of path|val)
    """
    res = dict()

    for root in roots:
        match = JEXP.find(root)
        n = 0
        hdrs = dict()

        for node in match:
            if not issubclass(node.value.__class__, (dict, list, UserDict)):
                # abc/def/ghi ->a.d.ghi
                fp = str(node.full_path).split('.')
                npre = (str(n)+'_') if num else ''
                if style == 'short':
                    # first char
                    key = ''.join((x[:1] + sep) for x in fp[:-1])
                elif style == 'full':
                    key = ''.join((x + sep) for x in fp[:-1])
                elif style == 'last2':
                    # last two segments
                    key = (fp[-2] + sep) if len(fp) > 1 else ''
                else:
                    key = ''
                hdrs[npre + key + fp[-1]] = fp if path_list else node.value
                n += 1
        res.update(hdrs)
    return res


@ lru_cache(maxsize=128)
def jexp(expr, *args, **kwds):
    return jex.parse(expr, *args, **kwds)


def getCacheInfo():
    info = {}
    for i in [jexp]:
        info[i] = i.cache_info()

    return info


def jsonPath(data, expr, val='simple', sep='/', indent=None, *args, **kwds):
    """ Make a JSONPath query on the data.

    :expr: JSONPath expression. Ref 'jsonpath_ng'

    :sep: '' or `None` for keeping `jsonpath_ng` format (e.g. `a.b.[3].d`; other string for substituting '.' to the given string, with '[' and ']' removed. Default is '/'.
    :val: 'context' for returning the `list` of `DatumInContext` of `find`; 'simple' (default) for list of simple types of values and summarizing `list` and `dict` values; other for a list of un-treated `DatumInContext.value`s; 'paths' for a list of paths only.
    :indent: for `json.dumps`.
    Returns
    -------
    If `val` is ```context```, return  the `list` of `DatumInContext` of `jsonpath_ng.ext.parse().find()`.
    Else return a `list` of `full_path`-`value` pairs from the output of `find().`
    * If `val` is ```simple```, only node values of simple types are kept, `list` and `dict` types will show as '<list> length' and '<dict> [keys [... [length]]]', respectively.
    * If `val` is ```full```, the values of returned `list`s are  un-treated `DatumInContext.value`s.
    """

    jsonpath_expression = jexp(expr, *args, **kwds)
    match = jsonpath_expression.find(data)
    if val == 'context':
        return match
    res = []
    for x in match:
        # make key
        key = str(x.full_path)
        if sep == '' or sep is None:
            pass
        else:
            key = key.replace('.', sep).replace('[', '').replace(']', '')
        if val == 'paths':
            res.append(key)
            continue
        # make value
        vc = x.value.__class__
        if val == 'simple':
            if issubclass(vc, (list)):
                value = f'<{vc.__name__}> {len(x.value)}'
            elif issubclass(vc, (dict)):
                n = 5
                ks = ', '.join(f'"{k}"' for k in
                               itertools.islice(x.value.keys(), n))
                l = len(x.value)
                if l > n:
                    ks += f'{ks}...({l})'
                value = f'<{vc.__name__}> {ks}'
            else:
                value = x.value
        elif val == 'full':
            value = x.value
        else:
            raise ValueError(
                'Invalid output type for jsonPath: %s' % str(val))
        res.append((key, value))
    return res



from unittest import TestCase
from fdi.httppool_app import create_app


class TestWelcome(TestCase):
    def setUp(self):
        self.app = create_app().test_client()

    def test_welcome(self):
        """
        Tests the route screen message
        """
        rv = self.app.get('/httppool/')

        # If we recalculate the hash on the block we should get the same result as we have stored
        self.assertEqual({"message": 'Hello World!'}, rv.get_json())
